/*
 * Xelfi
 */

package xelfi.awt;

import java.awt.*;

/**
 * A class that produces Gauge component.
 *
 * @version 2.1 1 Mar 1997
 * @author  Petr Hamernik
 */
public class Gauge extends Canvas {
	
	/** Horizontal gauge.
	 * @see #kind
	 */
	public final static int GAUGE_HORIZONTAL = 1;

	/** Vertical gauge.
	 * @see #kind
	 */
	public final static int GAUGE_VERTICAL = 2;

	/** Needle gauge.
	 * @see #kind
	 */
	public final static int GAUGE_NEEDLE = 3;
	
	/** Pie gauge.
	 * @see #kind
	 */
	public final static int GAUGE_PIE = 4;
	
	/** Text gauge.
	 * @see #kind
	 */
	public final static int GAUGE_TEXT = 5;
    
	/** Current kind of Gauge. For setting new kind use setKind method.
	 * @see #setKind
	 * @see #getKind
	 * @see #GAUGE_HORIZONTAL
	 * @see #GAUGE_VERTICAL
	 * @see #GAUGE_NEEDLE
	 * @see #GAUGE_PIE
	 * @see #GAUGE_TEXT
	 */
	protected int kind;

	/** When the Gauge is created the kind is set to this value. Default is GAUGE_HORIZONTAL.
	 * @see #GAUGE_HORIZONTAL
	 * @see #kind
	 */
	public static final int defaultKind = GAUGE_HORIZONTAL;

    /** Determine whether the text "xx%" is shown or hidden.
	 * @see #setShowText
	 * @see #getShowText
	 */
	protected boolean showText;

    /** Default value of showText. 
	 * @see #showText
	 */
	public static final boolean defaultShowText = true;

	/** Border color around the gauge. By default it is set to
	 *  defaultBorderColor.
	 * @see #setBorderColor
	 * @see #getBorderColor
	 * @see #defaultBorderColor
	 */
	protected Color marginColor = null;

	/** Default border color around the gauge. Currently black.
	 * @see #setBorderColor
	 * @see #getBorderColor
	 * @see #marginColor
	 */
    public static final Color defaultBorderColor = Color.black;

	/** Width of border around the gauge 
	 * @see #defaultBorderWidth
	 * @see #setBorderWidth
	 * @see #getBorderWidth
	 */	
    protected int margin = 0;
	
	/** Default width of 3D border around the gauge. Currently 1.
	 * @see #margin
	 * @see #setBorderWidth
	 * @see #getBorderWidth
	 */
    public static final int defaultBorderWidth = 1;
	
	/** None border around the gauge.
	 * @see #borderStyle
	 */
	public final static int BS_NONE = 0;

	/** Normal (non 3D) border around the gauge.
	 * @see #borderStyle
	 */
	public final static int BS_NORMAL = 1;
	
	/** 3D raised border around the gauge.
	 * @see #borderStyle
	 */
	public final static int BS_3D_RAISED = 2;
	
	/** 3D lowered border around the gauge.
	 * @see #borderStyle
	 */
	public final static int BS_3D_LOWERED = 3;
    
	/** Current border style around the gauge. For setting new bordeStyle use 
	 *  setBorderStyle method.
	 * @see #setBorderStyle
	 * @see #getBorderStyle
	 * @see #BS_NONE
	 * @see #BS_NORMAL
	 * @see #BS_3D_RAISED
	 * @see #BS_3D_LOWERED
	 */
	protected int borderStyle;

	/** When the Gauge is created the border style is set to this value. 
	 *  Default is BS_NORMAL.
	 * @see #BS_NORMAL
	 * @see #borderStyle
	 */
	public static final int defaultBorderStyle = BS_NORMAL;

    /** Current amount of the fill color. The gauge is painted by foreground color from
	 *  0% to the current value and by fillColor from the value to 100%. Use setFillColor
	 *  for setting this variable.
	 * @see #setFillColor
	 * @see #getFillColor
	 * @see #defaultFillColor
	 */
	protected Color fillColor = null;

	/** Default fill color. Currently lightGray.
	 * @see #setFillColor
	 * @see #getFillColor
	 * @see #fillColor
	 */
    public static final Color defaultFillColor = Color.lightGray;

	/** Current value of the gauge. Range of this variable is between 0 and 100.
	 *  Use setValue for setting new value.
	 * @see #setValue
	 * @see #getValue
	 */
	protected int value;

	/** Create new gauge component with all default values. */
	public Gauge() {
		setDefaults();
	}

	private void setDefaults() {
		kind = defaultKind;
		showText = defaultShowText;
    	marginColor = defaultBorderColor;
		margin = defaultBorderWidth;
		borderStyle = defaultBorderStyle;
    	fillColor = defaultFillColor;
		value = 0;
	}

	/** Sets new kind of gauge
	 * @param aKind New kind.
	 * @exception ArrayIndexOutOfBoundsException If the value was invalid.
	 * @see #kind
	 * @see #getKind
	 * @see #GAUGE_HORIZONTAL
	 * @see #GAUGE_VERTICAL
	 * @see #GAUGE_NEEDLE
	 * @see #GAUGE_PIE
	 * @see #GAUGE_TEXT
	 */
	public void setKind(int aKind) throws ArrayIndexOutOfBoundsException {
		if (aKind == kind)
			return;
		if ((aKind < 1) || (aKind > 5))
	        throw new ArrayIndexOutOfBoundsException();
		kind = aKind;
		repaint();
	}

	/** Gets current kind of gauge.
	 * @return Current kind.
	 * @see #kind
	 * @see #setKind
	 * @see #GAUGE_HORIZONTAL
	 * @see #GAUGE_VERTICAL
	 * @see #GAUGE_NEEDLE
	 * @see #GAUGE_PIE
	 * @see #GAUGE_TEXT
	 */
	public int getKind() {
		return kind;
	}

	/** Sets the border color.
	 * @param col New value of marginColor.
	 * @see #marginColor
	 * @see #defaultBorderColor
	 * @see #getBorderColor
	 */
    public void setBorderColor(Color col) {
        marginColor = col;
        repaint();
    }

	/** Gets the border color.
	 * @return Current border color.
	 * @see #marginColor
	 * @see #defaultBorderColor
	 * @see #setBorderColor
	 */
    public Color getBorderColor() {
        return marginColor;
    }

	/** Sets border width around the component.
	 * @param borderWidth New value of margin
	 * @see #margin
	 * @see #getBorderWidth
	 */
    public void setBorderWidth(int borderWidth) {
        margin = borderWidth;
        repaint();
    }

	/** Gets border width around the component.
	 * @see #margin
	 * @see #setBorderWidth
	 */
    public int getBorderWidth() {
        return(margin);
    }

	/** Sets the new value of showText variable.
	 * @param aShowText new value of showText.
	 * @see #showText
	 * @see #getShowText
	 */
    public void setShowText(boolean aShowText) {
        if (showText == aShowText)
			return;
		showText = aShowText;
        repaint();
    }

	/** Gets the current value of showText variable.
	 * @return Current value of showText
	 * @see #showText
	 * @see #setShowText
	 */
    public boolean getShowText() {
        return(showText);
    }

	/** Sets the new border style. Allowed values are all Gauge.BS_XXX constants.
	 * @param aBorderStyle new value of borderStyle.
	 * @see #borderStyle
	 * @see #getBorderStyle
	 * @see #BS_NONE
	 * @see #BS_NORMAL
	 * @see #BS_3D_RAISED
	 * @see #BS_3D_LOWERED
	 * @exception ArrayIndexOutOfBoundsException If the value was invalid.
	 */
    public void setBorderStyle(int aBorderStyle) throws ArrayIndexOutOfBoundsException {
		if (borderStyle == aBorderStyle)
			return;
		if ((aBorderStyle < 0) || (aBorderStyle > 3))
	        throw new ArrayIndexOutOfBoundsException();
		borderStyle = aBorderStyle;
        repaint();
    }

	/** Gets the current value of showText variable.
	 * @see #borderStyle
	 * @see #setBorderStyle
	 * @see #BS_NONE
	 * @see #BS_NORMAL
	 * @see #BS_3D_RAISED
	 * @see #BS_3D_LOWERED
	 */
    public int getBorderStyle() {
		return borderStyle;
    }

	/** Sets the fill color.
	 * @param col New value of fillColor.
	 * @see #fillColor
	 * @see #defaultFillColor
	 * @see #getFillColor
	 */
    public void setFillColor(Color col) {
		fillColor = col;
		repaint();
    }

	/** Gets the fill color.
	 * @return Current fill color.
	 * @see #fillColor
	 * @see #defaultFillColor
	 * @see #setFillColor
	 */
    public Color getFillColor() {
        return fillColor;
    }

	/** Sets new value. Allowed range is from 0 to 100 percent.
	 * @see #value
	 * @see #getValue
	 * @exception ArrayIndexOutOfBoundsException If the value was invalid.
	 */
	public void setValue(int aValue) throws IllegalArgumentException {
		if ((aValue < 0)||(aValue > 100))
			throw new IllegalArgumentException();
		value = aValue;
		repaint();
	}

	/** Gets current value.
	 * @return Current value.
	 * @see #value
	 * @see #setValue
	 */
	public int getValue() {
		return value;
	}

	public void paint(Graphics g) {
		Dimension d = size();

		int dif = (borderStyle == BS_NONE) ? 0 : margin;
		int x = dif;
		int y = dif;
		int w = d.width - 2*dif;
		int h = d.height - 2*dif;
		
		switch (kind) {
			case GAUGE_HORIZONTAL:	int w2 = (w * value) / 100;
									g.setColor(getForeground());
									g.fillRect(x,y,w2,h);
									g.setColor(fillColor);
									g.fillRect(x+w2,y,w-w2,h);
									break;
			case GAUGE_VERTICAL:	int h2 = (h * value) / 100;
									g.setColor(getForeground());
									g.fillRect(x,y,w,h2);
									g.setColor(fillColor);
									g.fillRect(x,y+h2,w,h-h2);
									break;
			case GAUGE_NEEDLE:		int angle = -(180 * value) / 100;
									g.setColor(fillColor);
									g.fillArc(x,y,w,h*2, angle, 180);
									g.setColor(getForeground());
									g.drawArc(x,y,w,h*2, angle, 180);
									g.fillArc(x,y,w,h*2,-180,angle);
									break;
			case GAUGE_PIE:			int angle1 = (360 * value)/100;
									if (value >= 50) {
										g.setColor(getForeground());
										g.fillOval(x,y,w,h);
										g.setColor(fillColor);
										g.fillArc(x,y,w,h,0,360-angle1);
									}
									else {
										g.setColor(fillColor);
										g.fillOval(x,y,w,h);
										g.setColor(getForeground());
										g.fillArc(x,y,w,h,0,-angle1);
									}
									g.setColor(getForeground());
									g.drawOval(x,y,w,h);
									break;
			case GAUGE_TEXT:		break;
		}
		paintBorder(g, d);
		if (showText)
			paintText(g, d);
	}

	private void paintBorder(Graphics g, Dimension d) {
		int x = 0;
		int y = 0;
		int w = d.width - 1;
		int h = d.height - 1;
		g.setColor(marginColor);
		switch (borderStyle) {
			case BS_NONE: 	break;
			case BS_NORMAL:	for (int i = 0; i < margin; i++) {
								g.drawRect(x, y, w, h);
								x++;
								y++;
								w -= 2;
								h -= 2;
							}
							break;
			case BS_3D_RAISED:
			case BS_3D_LOWERED:
							boolean isUp = (borderStyle == BS_3D_RAISED);
							for (int i = 0; i < margin; i++) {
								g.draw3DRect(x, y, w, h, isUp);
								x++;
								y++;
								w -= 2;
								h -= 2;
							}
							break;
		}
	}

	protected void paintText(Graphics g, Dimension d) {
		String s = new String(new StringBuffer(value+"%"));
		
		Font font = getFont();
		FontMetrics f = Toolkit.getDefaultToolkit().getFontMetrics(
				font==null ? new Font("Helvetica", Font.PLAIN, 12) : font);
		int w1 = f.stringWidth(s);
		int h1 = f.getHeight();
		g.setColor(getBackground());
		g.fillRect((d.width-w1)/2, (d.height-h1)/2, w1, h1);
		g.setColor(getForeground());
		g.drawString(s, (d.width-w1)/2, (d.height+h1)/2-2);
	}		

    public Dimension minimumSize() {
        return countSize();
    }

    public Dimension preferredSize() {
        return countSize();
    }

	private Dimension countSize() {
		Dimension d = new Dimension(100,50);
		if (borderStyle != BS_NONE) {
			d.width += 2*margin;
			d.height += 2*margin;
		}
		return d;
	}
}
