package xelfi.awt;

import java.awt.*;


/**
 * Owner draw list box in canvas, without scrollbars, but scrollable
 * through XelfiScrollable interface. Delivers almost all its 
 * functionality to the XFListView class.
	Last change:  DF    2 Feb 97    3:55 pm
*/
public class XFListViewCanvas extends Canvas implements XelfiScrollable {

	private XFListView		listView;

	// selection indexes
	private int 					startSelIndex;
	private int 					curSelIndex;


	/**
	 * Creates new empty signle sel. list
	*/
  public XFListViewCanvas () {
		this(false);
	}

	/**
	 * Creates new list with multiple selections 
	 * enabled/disabled as specified in param.
	*/
  public XFListViewCanvas (boolean multipleSelections) {
		listView = new XFListView(multipleSelections);
		initIndexes();
	}

  /**
	 * Adds the specified item to the end of scrolling list.
	 * Only reference to the item is added to the list.
	*/
	public void addItem (XelfiListViewElement item) {
		listView.addItem(item);
		repaint();
	}

	/**
	 * Adds the specified item to the specified index.
	 * Indexes are counted from 0 base.
	 * Only reference to the item is added to the list.
	*/
	public void addItem (XelfiListViewElement item, int index) {
		listView.addItem(item, index);
		repaint();
	}

	/**
	 * Returns true if this list allows multiple selections.
	*/
	public boolean allowsMultipleSelections () {
		return listView.allowsMultipleSelections();
	}

	/**
	 * Clears the list.
	*/
	public void clear () {
		listView.clear();
		initIndexes();
		repaint();
	}

	/**
	 * Returns the number of items in the list.
	*/
	public int countItems () {
		return listView.countItems();
	}

	/**
	 * Returns the number of selected items in the list.
	*/
	public int countSelectedItems () {
		return listView.countSelectedItems();
	}

	/**
	 * Delete an item from the list.<br>
	 * Item is specified by its index, 
	 * counted from 0 for first item in the list.
	 * Returns false if specified item couldn't be deleted;
	 * true otherwise.
	*/
  public boolean delItem (int index) {
		boolean wasSelected = isSelected(index);
		boolean result = listView.delItem(index);
		if (result) {
		  repaint();
			if (wasSelected)
				initIndexes();
		}
		return result;
	}

	/**
	 * Deselects the item at the specified index.
	 * Returns false if specified item hasn't been selected;
	 * true otherwise.
	*/
	public boolean deselect (int index) {
		boolean result = listView.deselect(index);
		if (result)	{
			initIndexes();
  	  repaint();
		}
		return result;
	}

	/**
	 * Gets the item associated with the specified index.
	*/
	public XelfiListViewElement getItem (int index) {
		return listView.getItem(index);
	}

	/**
	 * Get the selected item on the list or -1 if no item is selected.
	 * If multiple selections is set, returns index of 
	 * first selected item.
	*/
	public int getSelectedIndex () {
		return listView.getSelectedIndex();
	}

	/**
	 * Returns the selected indexes on the list.
	*/
	public int[] getSelectedIndexes () {
		return listView.getSelectedIndexes();
	}

	/**
	 * Returns the selected item on the list 
	 * or null if no item is selected.
	*/
	public XelfiListViewElement getSelectedItem () {
		return listView.getSelectedItem();
	}

	/**
	 * Returns the selected items on the list.
	*/
	public XelfiListViewElement[] getSelectedItems () {
		return listView.getSelectedItems();
	}

	/**
	 * Returns true if the item at the specified index has been selected; 
	 * false otherwise.
	*/
	public boolean isSelected (int index) {
		return listView.isSelected(index);
	}

	/**
	 * Replaces the item at the given index.
	*/
	public void replaceItem (XelfiListViewElement item, int index) {
		listView.replaceItem(item, index);
		repaint();
	}

	/**
	 * Selects the item at the specified index.
	*/
	public boolean select (int index) {
		boolean result = listView.select(index);
		if (result) {
			curSelIndex = index;
			startSelIndex = index;
			repaint();
		}
		return result;
	}

	/**
	 * Conditionally selects or unselects item at the specified index.
	*/
	public void select (int index, boolean selectItem) {
		boolean wasSelected = isSelected(index);
		listView.select(index, selectItem);
		if (selectItem) {
			curSelIndex = index;
			startSelIndex = index;
		} else 
		  if (wasSelected)
				initIndexes();
		repaint();
	}

	/**
	 * Enables/disables multiple selections for list.
	*/
	public boolean setMultipleSelections (boolean multiple) {
		boolean result = listView.setMultipleSelections(multiple);
		if (result) {
			initIndexes();
			repaint();
		}
		return result;
	}

	/**
	 * Returns index of an item which contains specified point.
	 * Point coordinates are relative to the getView() point.
	 * Returns -1 if no such item can be found.
	*/
	public int getIndexFromPoint (int x, int y) {
		return listView.getIndexFromPoint(x, y);
	}

	/**
	 * Returns absolute coordinates of the upper left point
	 * of the item at index.
	*/
	public Point getUpperLeft (int index) {
		return listView.getUpperLeft(index);
	}

	/**
	 * Makes na item with the specified index visible
	 * in the canvas.
	*/
	public void makeVisible (int index) {
		Point upperLeft = getUpperLeft(index);
		Point view = getView();
		Dimension itemSize = getItem(index).preferredSize();
		Dimension mySize = size();
		int yIndent = getYIndent();

		if (view.y>(upperLeft.y-yIndent)) {
			// item is invisible and scrolled above visible area...
			setView(view.x, upperLeft.y-yIndent);
			notifyScrollOwner();
		} else {
			if ((view.y+mySize.height)<(upperLeft.y+itemSize.height+yIndent)) {
				// item is invisible and scrolled below visible area...
				setView(view.x, upperLeft.y+itemSize.height+yIndent-mySize.height);
				notifyScrollOwner();
			}
		}
	}

	/**
	 * Access method. Returns horizontal indent.
	*/
	public int getYIndent () {
		return listView.getYIndent();
	}

	/**
	 * Access method. Returns vertical indent.
	*/
	public int getXIndent () {
		return listView.getXIndent();
	}

	/**
	 * Access method. Returns horizontal space between items.
	*/
	public int getSpaceBetween () {
		return listView.getSpaceBetween();
	}

	/**
	 * Access method. Sets horizontal indent.
	*/
	public void setYIndent (int yIndent) {
		listView.setYIndent(yIndent);
		repaint();
	}

	/**
	 * Access method. Sets vertical indent.
	*/
	public void setXIndent (int xIndent) {
		listView.setXIndent(xIndent);
		repaint();
	}

	/**
	 * Access method. Sets horizontal space between items.
	*/
	public void setSpaceBetween (int spaceBetween) {
		listView.setSpaceBetween(spaceBetween);
		repaint();
	}

	// Implementation of the interface XelfiScrollable follows
	/**
	 * Sets the upper left point of the view.
	*/
	public void setView (int x, int y) {
		listView.setView(x, y);
		repaint();
	}

	/**
	 * Returns upper left corner of the view.
	*/
	public Point getView () {
		return listView.getView();
	}

	/**
	 * Paints specified area, starting from getView() point
	 * with width and height as specified in params.
	*/
	public void paintArea (Graphics g, int width, int height) {
		listView.paintArea(g, width, height);
	}

	/**
	 * Paints specified area.
	*/
	public void paintArea (Graphics g, int x, int y, int width, int height,
												 boolean clearBackground) {
		listView.paintArea(g, x, y, width, height, clearBackground);
	}

	/**
	 * Returns visible size.
	 * Returns size of the canvas 
	 * - which should be exactly what is visible :-)
	*/
	public Dimension visibleSize () {
		return size();
	}

  /**
	 * Returns dimension needed for all scrollable
	 * object screen area to be shown.
	*/
	public Dimension wholeSize () {
		return listView.wholeSize();
	}

	/**
	 * Sets scrollable owner - object which holds scrollbars
	 * and should be notified when changes in coordinates of
	 * visible area are made.
	*/
	public void setScrollableOwner (XelfiScrollableOwner scrollableOwner) {
		listView.setScrollableOwner(scrollableOwner);
	}

	/**
	 * Gets scrollable owner - object which usually holds scrollbars
	*/
	public XelfiScrollableOwner getScrollableOwner () {
		return listView.getScrollableOwner();
	}


	/**
	 * Paints list. Calls paintArea with component's size.
	*/
	public void paint (Graphics g) {
		Dimension mySize = size();
		paintArea(g, mySize.width, mySize.height);
	}

	/**
	 * Notifies scrollable owner if size has been changed.
	*/
	public void reshape (int x, int y, int width, int height) {
		Dimension oldSize = size();
		super.reshape(x, y, width, height);
		// notify scrollable owner if size has been changed really
		if ((oldSize.width!=width) || (oldSize.height!=height)) {
			XelfiScrollableOwner scrollOwner = getScrollableOwner();
			if (scrollOwner!=null)
				scrollOwner.wholeSizeChanged();
		}
	}

	/**
	 * Handles key events.
	*/
	public boolean keyDown (Event e, int key) {
		boolean ShiftDown = e.shiftDown();
		boolean multi = allowsMultipleSelections();

		switch (key) {
			case Event.UP:
				if (ShiftDown && multi)
					shiftUpPressed();
				else
					upPressed();
				makeVisible(curSelIndex);
				break;
			case Event.DOWN:
				if (ShiftDown && multi)
					shiftDownPressed();
				else
					downPressed();
				makeVisible(curSelIndex);
				break;
		}
		return false;
	}

	/**
	 * Handles mouse events.
	*/
	public boolean mouseDown (Event e, int x, int y) {
		boolean ShiftDown = e.shiftDown();
		boolean ControlDown = e.controlDown();
		boolean multi = allowsMultipleSelections();

		if (ShiftDown && multi)
			shiftMousePressed(x, y);
    else {
			if (ControlDown && multi)
				controlMousePressed(x, y);
			else {
				int index = getIndexFromPoint(x, y);
				if (index>=0) {
					if (multi) 
					  listView.deselectAll();
					select(index);
				}
			}
		}
		return false;
	}

	private void notifyScrollOwner () {
		XelfiScrollableOwner scrollOwner = getScrollableOwner();
		if (scrollOwner!=null)
			scrollOwner.visibleAreaChanged();
	}

	private void initIndexes () {
		startSelIndex = -1;
		curSelIndex = -1;
	}

	private boolean noIndexes () {
		return curSelIndex==-1;
	}

	private boolean allowArrow (boolean down) {
		return ((down) && (curSelIndex<(listView.countItems()-1))) ||
           ((!down) && (curSelIndex>0));
	}

	private void initSelection (boolean down) {
		if (down) {
			startSelIndex = 0;
			curSelIndex = 0;
		} else {
			int itemCount = countItems();
			startSelIndex = itemCount-1;
			curSelIndex = itemCount-1;
		}
	}

	private void arrowPressed (boolean down) {
		if (noIndexes()) {
			initSelection(down);
			return;
		}
		// if no more to come, then return
		if (!allowArrow(down))
		  return;
		// deselect and select new
		int oldSelectCount =	countSelectedItems();
		listView.deselectAll();
		curSelIndex += down ? 1 : -1;
		startSelIndex = curSelIndex;
		listView.select(curSelIndex);
		if (oldSelectCount<=1) {
			// repaint only what that part that is needed
			Point curUL = getUpperLeft(curSelIndex);
			Graphics g = getGraphics();
			if (down) {
				Point oldUL = getUpperLeft(curSelIndex-1);
			 	paintArea(g, oldUL.x, oldUL.y,
			 	          size().width,
					        getItem(curSelIndex-1).preferredSize().height+
									getSpaceBetween()+1+
									getItem(curSelIndex).preferredSize().height,
									true);
			} else {
				Point oldUL = getUpperLeft(curSelIndex+1);
			 	paintArea(g, curUL.x, curUL.y,
			 		        size().width,
									getItem(curSelIndex).preferredSize().height+
									getSpaceBetween()+1+
									getItem(curSelIndex+1).preferredSize().height,
									true);
			}
		} else
			repaint();
	}


	private void upPressed () {
		arrowPressed(false);
	}

	private void shiftUpPressed () {
		if (noIndexes()) {
			initSelection(false);
			return;
		}
		if (!allowArrow(false))
		  return;
		if (startSelIndex<curSelIndex)
			listView.deselect(curSelIndex);
    curSelIndex--;
		listView.select(curSelIndex);
		repaint();
	}

	private void downPressed () {
		arrowPressed(true);
	}

	private void shiftDownPressed () {
		if (noIndexes()) {
			initSelection(true);
			return;
		}
		if (!allowArrow(true))
		  return;
		if (startSelIndex>curSelIndex)
			listView.deselect(curSelIndex);
    curSelIndex++;
		listView.select(curSelIndex);
		repaint();
	}

	private void fillWithSelection (int startIndex, int EndIndex) {
		listView.deselectAll();
		for (int i=Math.min(startIndex, EndIndex);
				 i<=Math.max(startIndex, EndIndex); i++) {
			listView.select(i);
		}
		repaint();
	}

	private void shiftMousePressed (int x, int y) {
		int index = getIndexFromPoint(x, y);
		if (index<0) return;
		if (noIndexes()) {
			select(index);
			return;
		}
		curSelIndex = index;
		fillWithSelection(curSelIndex, startSelIndex);
	}

	private void controlMousePressed (int x, int y) {
		int index = getIndexFromPoint(x, y);
		if (index<0) return;
		if (noIndexes()) {
			select(index);
			return;
		}
		curSelIndex = index;
		if (isSelected(index))
			listView.deselect(curSelIndex);
		else
 			listView.select(curSelIndex);
		repaint();
	}

}

