package xelfi.awt;

import java.awt.*;

import xelfi.util.Debug;

/**
 * A canvas for viewing hierarchically sorted strings in tree structure
 * Responds to key arrows and single mouse clicks by selecting proper item
 * Handles pulling down/up of subtrees (space or mouse click on icon).
 *
 * @version 1.02, 8 Dec 1996
 * @author Dafe
	Last change:  DF   17 Mar 97   12:24 pm
*/
public class XFTreeViewCanvas extends Canvas {

	// Id of notification messages
	public static final int			SELECTION_CHANGED = 4855+10;
	public static final int			DRAG_DROP = 4855+20;

	// an object which performs tree drawing and maintains the tree
	private XFTreeView					treeView;
	// translation - says where is the upper left point of the visible
	// part of the tree
	private int 								myX=0, myY=0;
	private XFTreeViewPanel			parent = null;
	// enables/disables event handling of user commands
	private boolean 						userInput;
	// enables/disables drag-drop feature
	private boolean 						dragDrop;
	private boolean 						wasDragged;
	private XFTreeNode					startDragNode, oldNode;

	// selection indexes
	private int									startSelIndex;
	private int									curSelIndex;

	/**
	 * Constructs new tree view canvas with specified 
	 * root element with no parent.<br>
	*/
	public XFTreeViewCanvas (Object rootElement) {
		this(rootElement, null);
	}

	/**
	 * Constructs new tree view canvas with specified 
	 * root element and with parent.<br>
	 * Parent will be notified about changes in properties which may affect
	 * scrollbars
	*/
	public XFTreeViewCanvas (Object rootElement, XFTreeViewPanel parent) {
		treeView = new XFTreeView(rootElement, this);
		this.parent = parent;
		userInput = true;

		dragDrop = false;
		wasDragged = false;
		startDragNode = null;
	}
	/**
	 * Inserts new view element after currently selected
	 * and selects it
	*/
	public void insertAfter (XelfiTreeViewElement element) {
		treeView.insertAfter(element);
		// repair selection indexes
		if (allowsMultipleSelections()) {
			curSelIndex = getSelectedIndex();
			startSelIndex = curSelIndex;
		}
		repaint();
		sendNotification(SELECTION_CHANGED);
	}

	/**
	 * Inserts new view element into currently selected
	 * and selects it
	*/
	public void insertInto (XelfiTreeViewElement element) {
		treeView.insertInto(element);
		if (allowsMultipleSelections()) {
			curSelIndex = getSelectedIndex();
			startSelIndex = curSelIndex;
		}
		repaint();
		sendNotification(SELECTION_CHANGED);
	}

	/**
	 * Inserts new view element after specified node.
	 * (on the same level in the tree).<br>
	 * Selection remains unchanged.
	*/
	public void insertAfter (XFTreeNode node, XelfiTreeViewElement element) {
		treeView.insertAfter(node, element);
		repaint();
		sendNotification(SELECTION_CHANGED);
	}

	/**
	 * Inserts new view element into specified node
	 * (item will be first in a childlist).<br>
	 * Selection remains unchanged.
	*/
	public void insertInto (XFTreeNode node, XelfiTreeViewElement element) {
		treeView.insertAfter(node, element);
		repaint();
		sendNotification(SELECTION_CHANGED);
	}

	/**
	 * Removes currently selected element,
	 * selects next one or previous one
	*/
	public void remove () {
		treeView.remove();
		repaint();
		// send notification 
		sendNotification(SELECTION_CHANGED);
	}

	/**
	 * Pulls currently selected element down or up, if it's possible.
	 * (it means if element can have children)
	*/
	public void pull (boolean state) {
		treeView.pull(state);
		repaint();
	}

	/**
	 * Toggles pull of currently selected element.
	 * Does nothing when multiple elements are selected.
	*/
	public void togglePull () {
		if (countSelectedItems()!=1) return;
		treeView.togglePull();
		repaint();
	}

	/**
	 * Selects next element in tree (from top to down)
	*/
	public void next () {
		int oldIndex = treeView.getCurIndex();
		int selCount = treeView.countSelectedItems();
		boolean result;
		boolean multi	= allowsMultipleSelections();
		if (multi) {
			if (oldIndex==-1) {
				selectNode(0);
				result = true;
			} else {
				result = treeView.next(curSelIndex);
				startSelIndex = treeView.getSelectedIndex();
				curSelIndex = startSelIndex;
			}
		} else
			result = treeView.next();
		makeCurrentVisible();
		if (result==true) {
			if (selCount<=1) {
				Graphics g = getGraphics();
				treeView.paintItem(g, oldIndex, -myX, -myY, true);
				treeView.paintItem(g, treeView.getCurIndex(), -myX, -myY, true);
				sendNotification(SELECTION_CHANGED);
			} else 
				repaint();
		}
	}

	/**
	 * Selects previous element in tree (from top to down)
	 * if it is possible.
	*/
	public void previous () {
		int oldIndex = treeView.getCurIndex();
		int selCount = treeView.countSelectedItems();
		boolean result;
		boolean multi	= allowsMultipleSelections();
		if (multi) {
			if (oldIndex==-1) {
				selectNode(treeView.getCount()-1);
				result = true;
			}	else {
				result = treeView.previous(curSelIndex);
				startSelIndex = treeView.getSelectedIndex();
				curSelIndex = startSelIndex;
			}
		} else
			result = treeView.previous();
		makeCurrentVisible();
		if (result==true) {
			if (selCount<=1) {
				Graphics g = getGraphics();
				treeView.paintItem(g, oldIndex, -myX, -myY, true);
				treeView.paintItem(g, treeView.getCurIndex(), -myX, -myY, true);
				sendNotification(SELECTION_CHANGED);
			} else 
				repaint();
		}
	}

	/**
	 * Sets translation between upper left corner of canvas and
	 * upper left corner of the drawing of tree structure.<br>
	 * Forces repainting.
	*/
	public void setView (int myX, int myY) {
		this.myX = myX;
		this.myY = myY;
		repaint();
	}

	public int getViewX () {
		return myX;
	}

	public int getViewY () {
		return myY;
	}

	/**
	 * Returns tree view instance
	*/
	public XFTreeView getTreeView () {
		return treeView;
        }

  /**
   * Sets new tree.
	*/
  public void setTree (XFTree newTree) {
    treeView.setTree(newTree);
	  repaint();
  }

	/**
	 * Draws the tree.
	*/
	public void paint (Graphics g) {
		Dimension mySize = size();
		treeView.paint(g, -myX, -myY, mySize.width, mySize.height);
		notifyParent();
	}

	/**
	 * Enables or disables user input (event handling) to this canvas
	*/
	public void enableUserInput (boolean userInput) {
		this.userInput = userInput;
	}

	/**
	 * Returns state of user input.
	*/
	public boolean getUserInput () {
		return userInput;		
	}

	public boolean mouseDown (Event e, int x, int y) {
		boolean multi = allowsMultipleSelections();
		boolean controlPressed = e.controlDown();
		boolean shiftPressed = e.shiftDown();

		if (multi && shiftPressed) {
			shiftMousePressed(x, y);
		} else {
			if (multi && controlPressed) {
				controlMousePressed(x, y);
			} else {
				if (selectClicked(x, y)!=true) {
					togglePullClicked(x, y);
				}
			}
    }
    // request a focus
    Debug.debug(Debug.INSPECTOR, "mouse event incoming to the tree - requesting a focus...");
    requestFocus();
    // propagate all events
    return false;
	}

	/**
	 * Handles incoming key events 
	*/
	public boolean keyDown (Event e, int key) {
		boolean multi = allowsMultipleSelections();
		boolean shiftPressed = e.shiftDown();
//    Debug.debug(Debug.INSPECTOR, "tree keydown: multi is "+multi+" shift is "+shiftPressed);
		switch (key) {
			case Event.HOME:
				goToRoot();
				break;
			case Event.END:
				goToEnd();
				break;
			case Event.UP:
				if (multi && shiftPressed)
					shiftUpPressed();
				else
					previous();
				break;
			case Event.DOWN:
				if (multi && shiftPressed)
					shiftDownPressed();
				else
					next();
				break; 
			case 32:  // space
				togglePull();
				break;
		}
		// allow further processing
		return false;
	}

	/**
	 * Handles events - enables or disables them.
	*/
	public boolean handleEvent (Event e) {
		// handle incoming events only if input is enabled
		if (userInput==false) return false;

		return super.handleEvent(e);
	}

	/**
	 * Selects clicked element 
	*/
	public boolean selectClicked (int myX, int myY) {
		boolean result = false;
		XFTreeNode clicked = getNodeFromPoint(myX, myY);
		if (clicked!=null) {
			deselectAll();
			result = selectNode(clicked);
//			repaint();
			sendNotification(SELECTION_CHANGED);
		}
		return result;
	}

	/**
	 * Selects a node in tree view.
	 * Additional pulling down is done if necessary.
	*/
	public boolean selectAbsNode (XFTreeNode node) {
		boolean result = treeView.selectAbsNode(node);
		repaint();
		return result;
	}

	/**
	 * Returns currently selected element
	*/
	public XFTreeNode getCurNode () {
		return treeView.getCurNode();
	}

	/**
	 * Returns currently selected item
	*/
	public XelfiTreeViewElement getCurItem () {
		return treeView.getCurItem();
	}

	/**
	 * Returns index of currently selected node
	*/
	public int getCurIndex () {
		return treeView.getCurIndex();
	}

	/**
	 * Returns element at index position (can return only elements
	 * which are pulled down at this time, can not return elements
	 * which are hidden because its parent is pulled up)<br>
	 * Returns null on error.
	*/
	public XFTreeNode getNode (int index) {
		return treeView.getNode(index);
	}

	/**
	 * Returns count of elements in tree.<br>
	 * Doesn't count elements which are hidden by pulling.
	*/
	public int getCount () {
		return treeView.getCount();
	}

	/**
	 * Sets selection to node at index.<br>
	 * Returns false if index position doesn't exist
	*/
	public boolean selectNode (int index) {
		boolean result = treeView.selectNode(index);
		if (result) {
			treeView.paintItem(getGraphics(), index, -myX, -myY, true);
			if (allowsMultipleSelections()) {
				startSelIndex = index;
				curSelIndex = index;
			}
		}
		return result;
	}

	/**
	 * Sets selection to node(node mustn't be hidden by pulling)<br>
	 * Returns false if node can't be found
	*/
	public boolean selectNode (XFTreeNode node) {
		boolean result = treeView.selectNode(node);
		int index = getIndex(node);
		if (result) {
			treeView.paintItem(getGraphics(), index, -myX, -myY, true);
			if (allowsMultipleSelections()) {
				startSelIndex = index;
				curSelIndex = index;
			}
		}
		return result;
	}

	/**
	 * Returns node which contains specified point.<br>
	 * Returns null if there's no node containing the point.
	*/
	public XFTreeNode getNodeFromPoint (int myX, int myY) {
		return treeView.getNodeFromPoint(this.myX+myX, this.myY+myY);
	}

	/**
	 * Returns node which contains specified point.<br>
	 * Returns null if there's no node containing the point.
	*/
	public XFTreeNode getPullIconFromPoint (int myX, int myY) {
		return treeView.getPullIconFromPoint(this.myX+myX, this.myY+myY);
	}

	/**
	 * Returns bounds of element at specified index.
	 * (indexes counted from 0 and only pulled down are counted)
	*/
	public Rectangle getAbsElementBounds (int index) {
		return treeView.getElementBounds(index);
	}

	/**
	 * Returns bounds of element at specified index.
	 * (indexes counted from 0 and only pulled down are counted)
	 * Bounds are relative to the getView() point.
	*/
	public Rectangle getElementBounds (int index) {
		Rectangle absBounds = getAbsElementBounds(index);
		if (absBounds==null) return null;
		return new Rectangle(absBounds.x-myX, absBounds.y-myY, 
												 absBounds.width, absBounds.height);
	}

	/**
	 * Toggles pull of element which 
	 * node image contains point specified by coordinates x, y
	*/
	public boolean togglePullClicked (int x, int y) {
		XFTreeNode clicked = getPullIconFromPoint(x, y);
		if (clicked==null) return false;
		// deselects all if multiple
		if (allowsMultipleSelections())
			deselectAll();
		// selects element 
		selectNode(clicked);
		// toggles its pull
		togglePull();
		return true;
	}

	public void goToRoot () {
		if (treeView.getCurIndex()==0) return;
		selectNode(0);
		makeCurrentVisible();
		repaint();
		sendNotification(SELECTION_CHANGED);
	}

	public void goToEnd () {
		if (treeView.getCurIndex()>=treeView.getCount()-1) return;
		selectNode(treeView.getCount()-1);
		makeCurrentVisible();
		repaint();
		sendNotification(SELECTION_CHANGED);
	}

	public Dimension preferredSize () {
		return treeView.preferredSize();		
	}

  public Dimension minimumSize () {
    return treeView.minimumSize();
  }
    
	/**
	 * Sets closed folder image.<br>
	 * imageFile must be valid full path to the image file.
	 * When image loading fails, returns false
	*/
	public boolean setFolderImage (String imageFile) {
		boolean result = treeView.setFolderImage(imageFile);
		if (result)
			repaint();
		return result;
	}

	/**
	 * Sets closed folder image.<br>
   * When image loading fails, returns false
	*/
	public boolean setFolderImage (Image image) {
		boolean result = treeView.setFolderImage(image);
		if (result)
			repaint();
		return result;
	}

	/**
	 * Sets open folder image.<br>
	 * imageFile must be valid full path to the image file.
	 * When image loading fails, returns false
	*/
	public boolean setOpenFolderImage (String imageFile) {
		boolean result = treeView.setOpenFolderImage(imageFile);
		if (result)
			repaint();
		return result;
	}

	/**
	 * Sets open folder image.<br>
	 * When image loading fails, returns false
	*/
	public boolean setOpenFolderImage (Image image) {
		boolean result = treeView.setOpenFolderImage(image);
		if (result)
			repaint();
		return result;
	}

	/**
	 * Sets open folder image.<br>
	 * imageFile must be valid full path to the image file.
	 * When image loading fails, returns false
	*/
	public boolean setNodeImage (String imageFile) {
		boolean result = treeView.setNodeImage(imageFile);
		if (result)
			repaint();
		return result;
	}

	/**
	 * Sets open folder image.<br>
	 * When image loading fails, returns false
	*/
	public boolean setNodeImage (Image image) {
		boolean result = treeView.setNodeImage(image);
		if (result)
			repaint();
		return result;
	}

	/**
	 * Returns true if this list allows multiple selections.
	*/
	public boolean allowsMultipleSelections () {
		return treeView.allowsMultipleSelections();
	}

	/**
	 * Returns the number of selected items in the list.
	*/
	public int countSelectedItems () {
		return treeView.countSelectedItems();
	}

	/**
	 * Delete an item at index (only pulled down nodes) 
	 * from the list.<br>
	 * Returns false if specified item couldn't be deleted;
	 * true otherwise.
	*/
  public boolean delItem (int index) {
		boolean result = treeView.delItem(index);
		if (result)
			repaint();
		return result;
	}

	/**
	 * Deselects the item at the specified index.
	 * Returns false if specified item hasn't been selected;
	 * true otherwise.
	*/
	public boolean deselect (int index) {
		boolean result = treeView.deselect(index);
		if (result)
			treeView.paintItem(getGraphics(), index, -myX, -myY, true);
		return result;
	}

	/**
	 * Get the selected item on the list or -1 if no item is selected.
	 * If multiple selections is set, returns index of 
	 * first selected item.
	*/
	public int getSelectedIndex () {
		return treeView.getSelectedIndex();
	}

	/**
	 * Returns the selected indexes on the list.
	*/
	public int[] getSelectedIndexes () {
		return treeView.getSelectedIndexes();
	}

	/**
	 * Returns the selected items on the list.
	*/
	public XelfiTreeViewElement[] getSelectedItems () {
		return treeView.getSelectedItems();
	}

	/**
	 * Returns true if the item at the specified index has been selected; 
	 * false otherwise.
	*/
	public boolean isSelected (int index) {
		return treeView.isSelected(index);
	}

	/**
	 * Conditionally selects or unselects item at the specified index.
	*/
	public void select (int index, boolean selectItem) {
		treeView.select(index, selectItem);
		if (allowsMultipleSelections()) {
			startSelIndex = index;
			curSelIndex = index;
		}
		treeView.paintItem(getGraphics(), index, -myX, -myY, true);
	}

	/**
	 * Enables/disables multiple selections for list.
	*/
	public boolean setMultipleSelections (boolean multiple) {
		boolean result = treeView.setMultipleSelections(multiple);
		if (result)
			repaint();
		return result;
	}

	/**
	 * Refreshes content of treeview and element bounds.
	 * Needed when preferred size of some element inthe tree is changed.
	*/
	public void refresh () {
		treeView.refresh();
		repaint();
	}

	private void notifyParent () {
		if (parent!=null)
			parent.updateScrollbars();
	}

	public void makeCurrentVisible () {
		makeVisible(treeView.getCurIndex());
	}

	/**
	 * Makes visible an item specified by index.
	 * (performs neccesary scrolling up/down when needed)
	*/
	public void makeVisible (int index) {
		Rectangle elemBounds = treeView.getElementBounds(index);
		if (elemBounds==null) return;
		Rectangle myBounds = bounds();
		int diff = myX+myBounds.x-elemBounds.x;
		if (diff>0) setView(myX-diff, myY);
		diff = (elemBounds.x+elemBounds.width)-(myX+myBounds.x+myBounds.width);
		if (diff>0) setView(myX+diff, myY);
		diff = myY+myBounds.y-elemBounds.y;
		if (diff>0) setView(myX, myY-diff);
		diff = (elemBounds.y+elemBounds.height)-(myY+myBounds.y+myBounds.height);
		if (diff>0) setView(myX, myY+diff);
	}

	/**
	 * Returns index of node in tree. (only visible items)
	 * Returns -1 when no such node exists.
	*/
	public int getIndex (XFTreeNode node) {
		return treeView.getIndex(node);
	}

	/**
	 * Deselects all selected items.
	*/
	public void deselectAll () {
		treeView.deselectAll();
		repaint();
	}

	/**
	 * Checks if drag-drop feature is enabled
	*/
	public boolean allowsDragDrop () {
		return dragDrop;
	}

	/**
	 * Enables/disables drag-drop feature
	*/
	public void setDragDrop (boolean dragDrop) {
		this.dragDrop = dragDrop;
	}

	/**
	 * Handles mouse-drag operation
	*/
	public boolean mouseDrag (Event evt, int x, int y) {
		if (!dragDrop) return false;
		if (wasDragged==false) {
		  // first occurence of dragging
			wasDragged = true;
			startDragNode =	getNodeFromPoint(x, y);
		}
		if (y<0) {
			// need to scroll up
			scrollUp(-y);
			return true;
		}
		int myHeight = size().height;
		if (y>myHeight) {
			// need to scroll down
			scrollDown(y-myHeight);
			return true;
		}
		// select a node from current point
		XFTreeNode node = getNodeFromPoint(x, y);
		if (node!=null) {
			if ((oldNode!=null) && (oldNode!=node)) {
        deselect(getIndex(oldNode));
			}
			selectNode(node);
			oldNode = node;
		}
		return true;
	}

	/**
	 * Handles mouse up (possible end of drag-drop operation)
	*/
	public boolean mouseUp (Event evt, int x, int y) {
		if ((!dragDrop) || (!wasDragged)) return false;
		XFTreeNode targetNode = getNodeFromPoint(x, y);
    if ((targetNode!=null) && (startDragNode!=targetNode)) {
      boolean isContainer = ((XelfiTreeViewElement)targetNode.
                             getData()).canHaveChildren();
      XFTreeNode realTarget, previousItem;
      // if target is container insert into it, else after target node
      if (isContainer) {
        realTarget = targetNode;
        previousItem = null;
      } else {
        realTarget = targetNode.getParent();
        previousItem = targetNode;
      }
      if (moveItem(realTarget, startDragNode, previousItem)) {
				// notify parent (and give him source and dest)
				sendNotification(DRAG_DROP,
              new XFDragDropInfo(realTarget.getData(),
                                 startDragNode.getData(),
                                 previousItem==null ? null
                                     : previousItem.getData()));
			}
		}
		wasDragged = false;
		startDragNode = null;
		return true;
	}

	/**
	 * Moves the source node into destination node.
	 * Returns false when operation is not possible.
	*/
	public boolean moveItem (XFTreeNode destination,
												   XFTreeNode source, XFTreeNode after) {
		boolean result = treeView.moveItem(destination, source, after);
		if (result)
			repaint();
		return result;
	}

	private void scrollUp (int amount) {
		int curViewX = getViewX();
		int curViewY = getViewY();
		if (curViewY<=0) return;
		setView(curViewX, Math.max(0, curViewY-amount));
		repaint();
	}

	private void scrollDown (int amount) {
		int curViewX = getViewX();
		int curViewY = getViewY();
		int limit = preferredSize().height-size().height;
		if (curViewY>=limit) return;
		setView(curViewX,
						Math.min(limit, curViewY+amount));
		repaint();
	}

	/**
	 * Creates notification event and sends it to itself.
	 * Event goes up in the AWT component hierarchy through propagation
	 * mechanism.
	*/
	private void sendNotification (int eventId) {
		postEvent(new Event(this, eventId, null));		
	}

	/**
	 * Creates notification event with argument.
	*/
	private void sendNotification (int eventId, Object argument) {
		postEvent(new Event(this, eventId, argument));
	}


	private boolean allowArrow (boolean down) {
		return ((down) && (curSelIndex<(treeView.getCount()-1))) ||
           ((!down) && (curSelIndex>0));
	}

	private void shiftUpPressed () {
//		Debug.debug(Debug.INSPECTOR, "I'm here...");
		if (treeView.getSelectedIndex()==-1) {
			selectNode(treeView.getCount()-1);
			makeCurrentVisible();
			return;
		}
		if (!allowArrow(false))
		  return;
		if (startSelIndex<curSelIndex) {
			treeView.deselect(curSelIndex);
			treeView.paintItem(getGraphics(), curSelIndex, -myX, -myY, true);
		}
    curSelIndex--;
		treeView.selectNode(curSelIndex);
		treeView.paintItem(getGraphics(), curSelIndex, -myX, -myY, true);
		makeVisible(curSelIndex);
	}

	private void shiftDownPressed () {
		if (treeView.getSelectedIndex()==-1) {
			selectNode(0);
			makeCurrentVisible();
			return;
		}
		if (!allowArrow(true))
		  return;
		if (startSelIndex>curSelIndex) {
			treeView.deselect(curSelIndex);
			treeView.paintItem(getGraphics(), curSelIndex, -myX, -myY, true);
		}
    curSelIndex++;
		treeView.selectNode(curSelIndex);
		treeView.paintItem(getGraphics(), curSelIndex, -myX, -myY, true);
		makeVisible(curSelIndex);
	}

	private void fillWithSelection (int startIndex, int EndIndex) {
		treeView.deselectAll();
		for (int i=Math.min(startIndex, EndIndex);
				 i<=Math.max(startIndex, EndIndex); i++) {
			treeView.selectNode(i);
		}
		repaint();
	}

	private void shiftMousePressed (int x, int y) {
		XFTreeNode node = getNodeFromPoint(x, y);
		if (node==null) return;
		int index = getIndex(node);
		if (treeView.getSelectedIndex()==-1) {
			selectNode(index);
			return;
		}
		curSelIndex = index;
		fillWithSelection(curSelIndex, startSelIndex);
	}

	private void controlMousePressed (int x, int y) {
		XFTreeNode node = getNodeFromPoint(x, y);
		if (node==null) return;
		int index = getIndex(node);
		if (treeView.getSelectedIndex()==-1) {
			selectNode(index);
			return;
		}
		curSelIndex = index;
		treeView.select(curSelIndex, !isSelected(index));
		treeView.paintItem(getGraphics(), curSelIndex, -this.myX, -this.myY, true);
	}

}
