/**
* Xelfi
*/

package xelfi.design.forms;

import java.io.*;
import xelfi.design.components.XelfiDesignComponent;
import xelfi.design.components.XFPropertyIndexOutOfBoundsException;
import xelfi.editor.Source;

/**
 * XelfiDesignForm interface is a common interface for forms.
 * In contrast to XelfiDesignComponent and its descendants it is not possible to make
 * custom forms by implementing this interface.
 * The set of forms is closed to Frame (+Application), Dialog and Applet
 * @version 	1.62, 26 Feb 1997
 * from 1.10 inherits from XelfiDesignInspectable
 * from 1.20 has method getDefinedEvents()
 * from 1.30 has methods for event management
 * from 1.40 has method getSource()
 * from 1.50 has methods set/getPackage()
 * from 1.60 extends DocumentHandler
 * from 1.61 has methods setTemplateName, getTemplateName
 * @finished
 * @author 	Ian Formanek
 */
public interface XelfiDesignForm extends XelfiDesignInspectable {

	/**
	 * This method returns package in which is this form.
	 * @return 		package of this form
	 * @see setPackage
	 */
	public String getPackage(); 

	/**
	 * This method sets package of this form.
	 * @param	aPackageName 	new package of the form
	 * @see getPackage
	 */
	public void setPackage(String aPackageName); 

	/**
	 * This method returns Source that contains java source code for this form.
	 * @return 		Source for this form
	 */
	public Source getSource(); 

	/**
	 * This method sets given name to given component - it checks if no other component has
	 * this name.
	 * @return 		true if name was set O.K., false otherwise
	 */
	public boolean setComponentName(XelfiDesignComponent comp, String newName);

	////////////
	// events //
	////////////

	/**
	 * This method returns names of all event handlers currently defined in this form
	 * (these are methods that are called on specific event - user assigns and specifies
	 *  their name in the events page of ComponentInspector).
	 * It is (probably) neeeded by ComponentInspector only.
	 * @return 		array of string names of event hanling methods
	 */
	public String[] getDefinedEvents();

	/**
	 * This method is called by ComponentInspector when user assigns an event of given name 
	 * to given event on given component. The name of new event could be either an existing 
	 * handler or a new one.
	 * @param 	xdc		Component which event to assign
	 * @param 	index		Index of event to assign
	 * @param 	evName	Name of new event handler
	 */
	public void assignEvent(XelfiDesignComponent xdc, int index, String evName)
	throws XFPropertyIndexOutOfBoundsException;

	/**
	 * This method is called by ComponentInspector to show given event in editor.
	 * It is used for instance when user adds new event handler in ComponentInspector.
	 * @param 	evName	Name of event handler to show
	 */
	public void focusEvent(String evName);
	
	/**
	 * This method is called by ComponentInspector when user renames an event from given old name 
	 * to given new name.
	 * @param 	oldName		Old event name
	 * @param 	newName		New event name
	 */
	public void renameEvent(String oldName, String newName);

	/**
	 * This method is called by ComponentInspector when user removes an event of given event 
	 * on given component. Form decreases reference count for this name and if it becomes 0, asks
	 * user whether to delete the event handler.
	 * @param 	xdc		Component which event to assign
	 * @param 	index		Index of event to assign
	 * @param 	evName	Name of new event handler
	 */
	public void releaseEvent(XelfiDesignComponent xdc, int index)
	throws XFPropertyIndexOutOfBoundsException;

	//////////////
	// template //
	//////////////

	/**
	 * This method returns name of the template that this form originated from.
	 * @return 		package of this form
	 * @see setTemplateName
	 */
	public String getTemplateName(); 

	/**
	 * This method sets name of the template that this form originated from.
	 * It is not necessary to call it for plain forms (Application, Frame, Dialog, Applet),
	 * so as these forms have the name as default. In the other hand for forms
	 * that originated from a template file (stored form) it is necessary to call this
	 * method with appropriate name after it is created.
	 * @param	aTemplateName 	name of the template that this form originated from
	 * @see getTemplateName
	 */
	public void setTemplateName(String aTemplateName); 

	/////////////////////
	// code generation //
	/////////////////////
	
	/**
	 * This method is called whenever some code of given component must be regenerated
	 * (i.e. generated again because it changed).
	 * @param xdc		component which code is to be regenerated
	 * @param what	bitmap (of constants defined in XFDesignConstants class) which tells 
	 *						which sections to regenrate
	 */
	public void regenerateCode(XelfiDesignComponent xdc, int what);

	/////////////////
	// persistence //
	/////////////////
	
	/**
	 * Method which loads this form from stream (i.e. its properties and all its components).
	 * @param stream		stream to load form from
	 */
	public void loadFromStream(DataInput stream) throws
		IOException;
	
	/**
	 * Method which saves this form to stream.
	 * @param stream		stream to save form to
	 */
	public void saveToStream(DataOutput stream) throws
		IOException;

}