/**
* Xelfi programming environment
*/

package xelfi.editor;
import java.io.*;
import java.util.Enumeration;
/**
* Block is a pair of position one at begining and one at end
*
* @version 1.0
* @author Jaroslav Tulach
*/

public class Block extends Object {
  private static final String ET_DIFF = "Positions are in different documents";
  private static final String ET_LESS = "Positions are in bad order";
  private static final String ET_NOTCONTAINED = "Position is not in the block";

  /** Constant for comparing position of Position to Block.
  * Position is in different document.
  */
  public static final byte DIFFERENT = (byte)0x00;
  
  /** Constant for comparing position of Position to Block.
  * Position is before the block.
  */
  public static final byte BEFORE = (byte)0x01;
  
  /** Constant for comparing position of Position to Block.
  * Position is at the begin of the block.
  */
  public static final byte BEGIN = (byte)0x02;

  /** Constant for comparing position of Position to Block.
  * Position is inside the block..
  */
  public static final byte INSIDE = (byte)0x04;

  /** Constant for comparing position of Position to Block.
  * Position is at the end of the block.
  */
  public static final byte END = (byte)0x08;
  
  /** Constant for comparing position of Position to Block.
  * Position is after the block.
  */
  public static final byte AFTER = (byte)0x10;

  
  /** begin of text (first position) */
  Position begin;
  /** end of text (last position) */
  Position end;

  /** creates empty block, used by Document 
  * @see Document
  */
  protected Block () {
    begin = end = null;
  };
  
  /** creates new block that begin at b and ends at e
  * @param b begin of block
  * @param e end of block
  * @exception InvalidPositionException is raised when position are
  *   in different documents or e is before b
  */
  public Block (Position b, Position e) throws InvalidPositionException {
    setPositions (b, e);
  };
  
  /** Creates new block on specified position and inserts text to the block.
  * Null info is sent to allowInsert
  * @param p the position to insert text
  * @param str a text to insert
  * @return the new block
  * @exception NotAllowedException if the text is not allowed to be inserted
  */
  public static Block text (Position p, String str) throws NotAllowedException {
    return text (p, str, null);
  };
    
  /** Creates new block on specified position and inserts text to the block
  * @param p the position to insert text
  * @param str a text to insert
  * @param info the info to be sent to allowInsert
  * @return the new block
  * @exception NotAllowedException if the text is not allowed to be inserted
  */
  public static Block text (Position p, String str, Object info) throws NotAllowedException {
    try {
      return new Block (p.insert (str, info), p);
    } catch (InvalidPositionException ex) {
      // never
      return null;
    }
  };
  
  /** creates block that contains whole line with position p
  * @param p the Position
  * @return the new block
  */
  public static Block line (Position p) {
    try {
      return new Block (p.beginOfLine (), p.endOfLine ());
    } catch (InvalidPositionException ex) {
      // never
      return null;
    }
  };

  /** Creates new block. The first position is a and last b or vice versa.
  * @param a start/end the position
  * @param b start/end the position
  * @return the new block
  * @exception InvalidPositionException if a and are in different documents
  */
  public static Block between (Position a, Position b) throws InvalidPositionException {
    switch (a.compare (b)) {
    case Position.LESS:
    case Position.EQUAL:
      return new Block (a, b);
    case Position.GREATER:
      return new Block (b, a);
    default:
      throw new InvalidPositionException (ET_DIFF);
    }
  };
  
  /** sets positions to new values
  * @param b beginning of block
  * @param e end of block
  * @exception InvalidPositionException is raised when position are in
  *   different document or e is before b
  */
  private final void setPositions (Position b, Position e)
  throws InvalidPositionException {
    switch (b.compare (e)) {
      case Position.DIFFERENT:
        throw new InvalidPositionException (ET_DIFF);
      case Position.EQUAL:
      case Position.LESS:
        begin = b;
        end = e;
        break;
      default:
        throw new InvalidPositionException (ET_LESS);
    };
  };
  
  /** @return the first position in document */
  public Position getBegin () {
    return begin;
  };
  
  /** @return the last position in document */
  public Position getEnd () {
    return end;
  };
  
  /** For every position in block from begining till end calls
  * accept (Pos) till accept returns true

  * @param ac the acceptor
  * @param dir the direction. true means forward
  *
  * @return first position that accept returned true
  *    or null if accept always returned false (for all positions in ac)
  * @exception EndOfTextException if the end of block is reached
  */
  public final Position find (Acceptor ac, boolean dir) throws EndOfTextException {
    if (dir) {
      return getBegin ().find (ac, getEnd (), true);
    } else {
      return getEnd ().find (ac, getBegin (), false);
    }
  };
  
  /** For every position in block from begining till end calls
  * accept (Pos) till accept returns true

  * @param ac the acceptor
  * @return first position that accept returned true
  *    or null if accept always returned false (for all positions in ac)
  * @exception EndOfTextException if the end of block is reached
  */
  public final Position find (Acceptor ac) throws EndOfTextException {
    return find (ac, true);
  };
  
  /** Deletes whole block.
  * @param info the info to send to ActionControllers
  * @exception NotAllowedException if the action if forbiden
  */
  public final void delete (Object info) throws NotAllowedException {
    getBegin ().deleteTill (getEnd (), info);
  };

  /** Deletes whole block. The info is set to null.
  * @exception NotAllowedException if the action if forbiden
  */
  public final void delete () throws NotAllowedException {
    delete (null);
  };

  /** Deletes block and sets its content to new value.
  *
  * @warning In this version a strange thing can happen. Because the
  * insert is done first it can be allowed but the delete not. It
  * means that new text is inserted and the old does not deleted.
  *
  * @param str the new content of block
  * @param info the info to be sent do ActionControllers
  * @exception NotAllowedException if the action if forbiden
  */
  public final void set (String str, Object info) throws NotAllowedException {
    Position b = getBegin ();
    try {
      b.getDocument ().redrawLock ();
      b.deleteTill (b = getEnd ().insert (str, info), info);
      begin = b;
    } finally {
      b.getDocument ().redrawUnlock ();
    };
  };

  /** Deletes block and sets its content to new value.
  * Info is set to null.
  *
  * @warning In this version a strange thing can happen. Because the
  * insert is done first it can be allowed but the delete not. It
  * means that new text is inserted and the old does not deleted.
  *
  * @param str the new content of block
  * @exception NotAllowedException if the action if forbiden
  */
  public final void set (String str) throws NotAllowedException {
    set (str, null);
  };

  /** Compares position to the block.
  
  * @return  one of the constants BEFORE, BEGIN, INSIDE, END, AFTER
  * @see Block#BEFORE
  * @see Block#BEGIN
  * @see Block#INSIDE
  * @see Block#END
  * @see Block#AFTER
  * @see Block#DIFFERENT
  * @warning If the block is empty and the position is on the block then the @see Block#END
  * will be returned.
  */
  public final byte compare (Position p) {

    byte res;
    Position b = getBegin ();
    
    if (p.getDocument () != b.getDocument ()) {
      res = DIFFERENT;
    } else {
      Position e = getEnd ();
      if (p.equals (e)) {
        res = END;
      } else {
        if (p.equals (b)) {
          res = BEGIN;
        } else {
          if (p.less (b)) {
            res = BEFORE;
          } else {
            if (p.greater (e)) {
              res = AFTER;
            } else {
              res = INSIDE;
            };
          };
        };
      };
    };
    return res;
  };
  
  /** Tests if position is in the block.
  * Warning: begining of block is not in block (because insert at this position
  * will add chars before the beginning)
  * End of block is in block (insert always inserts before)
  * @param pos the position
  * @return true or false
  */
  public final boolean containsOrEnds (Position pos) {
    return (compare (pos) & (INSIDE | END)) != 0;
  };

  /** Tests if position is in the block or at the begining of it.
  * @param pos the postion
  * @return true or false
  */
  public boolean containsOrBegins (Position pos) {
    return (compare (pos) & (INSIDE | BEGIN)) != 0;
  };
  
  /** tests if position is in the block but not in the beginning or end.
  * @param p the position
  * @return true or false
  */
  public boolean contains (Position pos) {
    return (compare (pos) & INSIDE) != 0;
  };
  
  /** tests if position is in the block or at the beginning or end.
  * @param p the position
  * @return true or false
  */
  public boolean containsOrBeginsOrEnds (Position pos) {
    return (compare (pos) & (INSIDE | END | BEGIN)) != 0;
  };

  /** converts block to string
  * @return the string that is between position in this block
  */
  public String toString () {
    return getBegin ().toString (getEnd ());
  };
  
  /** saves block to stream
  * @param s the stream to save this file to
  * @exception IOException if there is an IO error
  */
  public void save (DataOutput s) throws IOException {
    try {
      Saver saver = new Saver (s);
      find (saver);
      // Only if an error occures the end of find is reached.
      // So we should throw an exception.
      throw saver.exception;
    } catch (EndOfTextException ex) {
      // whole block has been saved
    };
  };
  
  /** Is a block empty?
  * @return true if the block is empty
  */
  public boolean isEmpty () {
    Position b = getBegin ();
    Position e = getEnd ();
    return b.equals (e);
  };
  
  /** Returns block that is contained by both, this and b.
  * @param b the second block
  */
  public Block intersection (Block b) {
    
    Position a = getBegin ().less (b.getBegin ()) ? b.getBegin () : getBegin ();
    Position c = getEnd ().less (b.getEnd ()) ? getEnd () : b.getEnd ();
    try {
      return new Block (a, c);
    } catch (InvalidPositionException ex) {
      try {
        return new Block (a, a);
      } catch (InvalidPositionException ex1) {
        return null;
        // never happens
      }
    }
  };
  
  /** Returns block that covers both this and b. If these blocks are not
  * on the same document the empty block is returned.
  * @param b the second block
  */
  public Block union (Block b) {
    Position a = getBegin ().less (b.getBegin ()) ? getBegin () : b.getBegin ();
    Position c = getEnd ().less (b.getEnd ()) ? b.getEnd () : getEnd ();
    try {
      return new Block (a, c);
    } catch (InvalidPositionException ex) {
      try {
        return new Block (a, a);
      } catch (InvalidPositionException ex1) {
        return null;
        // never happens
      }
    }
  };

  /** Loads block from a stream.
  * @param p the position to load at
  * @param s the input stream
  * @return new block
  * @exception IOException if an error occures on read
  * @exception NotAllowedException if insert is not allowed
  */
  public static Block load (Position p, DataInput s)
  throws IOException, NotAllowedException {
    return load (p, s, null);
  };

  /** Loads block from a stream.
  * @param p the position to load at
  * @param s the input stream
  * @param info the info
  * @return new block
  * @exception IOException if an error occures on read
  * @exception NotAllowedException if insert is not allowed
  */
  public static Block load (Position p, DataInput s, Object info)
  throws IOException, NotAllowedException {
    try {
      return new Block (p.insert (s, info), p);
    } catch (InvalidPositionException ex) {
      return null;
      // never
    }
  };

  /** Converts content of block to input stream.
  * @return the input stream
  * @exception IOException if an I/O error occures
  */
  public DocumentInputStream getInputStream () throws IOException {
    return new DocumentInputStream (this);
  };
  
  /** Sends a message to block's document with a request to redraw. If a
  * part of the document is visible, than it should be repainted.
  */
  public void redraw () {
    getBegin ().getDocument ().redrawBlock (this);
  };
  
  /** Returns enumeration of every line begining in the block. First element
  * is begining of the block. Next one is added for each new line begining
  * till the end of block is reached.
  */
  public Enumeration lines () {
    return new LineEnumeration (this);
  };
  
  /** Counts number of lines in the block. If there is no new line then one
  * is returned.
  * @return the number of lines
  */
  public final int numberOfLines () {
    try {
      return lineToInt (getEnd ());
    } catch (InvalidPositionException ex) {
      // never happens
      return 0;
    }
  };
  
  /** Counts number of characters in the block.
  * The character at the end of the block is not counted. That is why if
  * the block begin equals block end then zero is returned.
  * @return number of characters
  */
  public final int numberOfChars () {
    Line first = Position.line (getBegin ().index);
    Line last = Position.line (getEnd ().index);
    int cnt = 0;
    while (first != last) {
      cnt += first.text.length ();
      first = first.nextLine;
    };
    return cnt;
  };

  /** Returns number of line on that specified position is. Counts from
  * begining of the block. If position is on first line returns zero.
  * @param the position to test
  * @return line number
  * @exception InvalidPositionException if p is before begin or after end
  */
  public int lineToInt (Position p) throws InvalidPositionException {
    if (containsOrBeginsOrEnds (p)) {
      int first = getBegin ().lineNumber ();
      int last = p.lineNumber ();
      return last - first;
    } else {
      throw new InvalidPositionException (ET_NOTCONTAINED);
    }
  };

  /** Returns begin of line n lines down from begin of block.
  * The line cannot be out of block.
  * @param n number of lines down
  * @return the specified position
  * @exception EndOfTextException if end of block is reached
  */
  public final Position goToLine (int i) throws EndOfTextException {
    Position p = getBegin ().beginOfLine ().down (i);
    if (p.greater (getEnd ())) {
      throw new EndOfTextException ();
    };
    return p;
  };
};



