/**
* Xelfi programming environment
*/

package xelfi.editor;
import java.util.Vector;
import java.util.NoSuchElementException;
import java.util.Enumeration;
import java.io.IOException;
import java.io.DataInput;
/**
* One text document. Allows creation and access to begin and end.
*
* @version 1.0
* @author Jaroslav Tulach
*/

public class Document extends Block implements
  ActionController,
  ActionPreListener,
  ActionPostListener
{

  /** arrays for listeners */
  private Vector ac = new Vector ();
  private Vector pre = new Vector ();
  private Vector post = new Vector ();
  private Vector redraw = new Vector ();
  private Vector doc = new Vector ();
  
  private int locks;
  private Line first;

  private boolean modified = false;

  /** Creates new empty document.
  */
  public Document () {
    first = new Line ();
    first.doc = this;
    begin = end = new Position (first.beginOfLine ());
  };
  
  /** creates new document by loading it from a stream
  * @param s the stream to load from
  * @exception IOException if an error occured during loading
  */
  public Document (DataInput s) throws IOException {
    this ();
    try {
      Block.load (begin, s);
      // delete new lines at the end of document
    } catch (NotAllowedException e) {
      // cannot happen because no AccessControler has been set yet
    };
  };

  /** Gets the first position in document. This position is valid only in time
  * of return. Than another thread can add new text before.
  * @return the first position in document
  */
  public Position getBegin () {
    return new Position (first.beginOfLine ());
  };

  /** Gets the last position in document. This position is valid only temporaly.
  * Someone can insert anything after this one
  */
  public Position getEnd () {
    end = end.beginOfLine ();
    try {
      for (;;) {
        end = end.down ();
      }
    } catch (EndOfTextException ex) {
    };
    return end.endOfLine ();
  };

  /**
  * Adds new next action controler to this document.
  * @param a the controler
  * @see ActionController
  */
  public final void addActionController (ActionController a) {
    ac.addElement (a);
  };

  /**
  * Adds new action pre listener to this document.
  * @param a the listener
  * @see ActionPreListener
  */
  public final void addActionPreListener (ActionPreListener a) {
    pre.addElement (a);
  };

  /**
  * Adds new action post listener to this document.
  * @param a the listener
  * @see ActionPostListener
  */
  public final void addActionPostListener (ActionPostListener a) {
    post.addElement (a);
  };

  /**
  * Adds new redraw listener to this document.
  * @param a the listener
  * @see RedrawListener
  */
  public final void addRedrawListener (RedrawListener a) {
    redraw.addElement (a);
  };

  /**
  * Adds new document listener to this document.
  * @param a the listener
  * @see DocumentListener
  */
  public final void addDocumentListener (DocumentListener a) {
    doc.addElement (a);
  };

  /** Only for internal use. Adds controler but before other ones.
  * This is needed by EditorImage to be informed before anyone reacts to the
  * notifyDelete or notifyInsert messages.
  * @param a the controler
  */
  final void prependActionPostListener (ActionPostListener a) {
    post.insertElementAt (a, 0);
  };
  
  /**
  * Removes action controler from this document.
  * @param a the controler to remove
  * @return false if a is not a controler for this document
  */
  public final boolean removeActionController (ActionController a) {
    return ac.removeElement (a);
  };

  /**
  * Removes action pre listener from this document.
  * @param a the controler to remove
  * @return false if a is not a listener for this document
  */
  public final boolean removeActionPreListener (ActionPreListener a) {
    return pre.removeElement (a);
  };

  /**
  * Removes action post listener from this document.
  * @param a the controler to remove
  * @return false if a is not a listener for this document
  */
  public final boolean removeActionPostListener (ActionPostListener a) {
    return post.removeElement (a);
  };

  /**
  * Removes redraw listener from this document.
  * @param a the controler to remove
  * @return false if a is not a listener for this document
  */
  public final boolean removeRedrawListener (RedrawListener a) {
    return redraw.removeElement (a);
  };

  /**
  * Removes document listener from this document.
  * @param a the controler to remove
  * @return false if a is not a listener for this document
  */
  public final boolean removeDocumentListener (DocumentListener a) {
    return doc.removeElement (a);
  };

  /** This method is called before every insert to any position in this
  * document. For multiple inserts (of arrays or Strings) is called only
  * once at begining.
  *
  *
  * @param p the position method insert will insert to
  * @param info the info
  * @exception NotAllowedException if this insert is not allowed
  */
  public void allowInsert (Position p, Object info) throws NotAllowedException {
    try {
      Enumeration e = ac.elements ();
      for (;;) {
        ((ActionController)(e.nextElement ())).allowInsert (p, info);
      }
    } catch (NoSuchElementException x) {
    }
  };
  
  /** This method is called before every delete at any position in this
  * document. Is called for every position that the delete will remove
  * (even for multiple deletes).
  *
  * @param p the position method delete will delete at
  * @param info the info
  * @exception NotAllowedException if this delete is not allowed
  */
  public void allowDelete (Position p, Object info) throws NotAllowedException {
    try {
      Enumeration e = ac.elements ();
      for (;;) {
        ((ActionController)(e.nextElement ())).allowDelete (p, info);
      }
    } catch (NoSuchElementException x) {
    }
  }

  /** This method is called before every delete of a block (more characters).
  * @param from the first position to delete
  * @param to the last (excluded) position
  * @param info the info
  * @exception NotAllowedException if this delete is not allowed
  */
  public void allowMultiDelete (Position from, Position to, Object info)
  throws NotAllowedException {
    try {
      Enumeration e = ac.elements ();
      for (;;) {
        ((ActionController)(e.nextElement ())).allowMultiDelete (from, to, info);
      }
    } catch (NoSuchElementException x) {
    }
  }

  /** Notifies before insert.
  * @param p position of insert
  * @param s the string to insert
  */
  public void beforeInsert (Position p, String s) {
    Enumeration e = pre.elements ();
    while (e.hasMoreElements ()) {
      ((ActionPreListener)(e.nextElement ())).beforeInsert (p, s);
    };
  };
  
  /** Notifies before delete of one character.
  * @param p position of delete
  */
  public void beforeDelete (Position p) {
    Enumeration e = pre.elements ();
    while (e.hasMoreElements ()) {
      ((ActionPreListener)(e.nextElement ())).beforeDelete (p);
    };
  };

  /** Notifies before multi delete.
  * @param from the first (included)
  * @param to the last (excluded)
  */
  public void beforeMultiDelete (Position from, Position to) {
    Enumeration e = pre.elements ();
    while (e.hasMoreElements ()) {
      ((ActionPreListener)(e.nextElement ())).beforeMultiDelete (from, to);
    };
  };

  /** Notifies after an insert has been done.
  */
  public void notifyInsert (Position n, Position p) {
    try {
      Enumeration e = post.elements ();
      for (;;) {
        ((ActionPostListener)(e.nextElement ())).notifyInsert (n, p);
      }
    } catch (NoSuchElementException x) {
    }
    setModified (true);
  };

  /** Notifies after a delete has been done
  */
  public void notifyDelete (Position p, String s) {
    try {
      Enumeration e = post.elements ();
      for (;;) {
        ActionPostListener ac = (ActionPostListener) e.nextElement ();
        ac.notifyDelete (p, s);
      }
    } catch (NoSuchElementException x) {
    }
    setModified (true);
  }

  /** Returns number of line on that specified position is. Counts from
  * begining of the block. If position is on first line returns zero.
  * @param the position to test
  * @return line number
  * @exception InvalidPositionException if p is before begin or after end
  */
  public int lineToInt (Position p) throws InvalidPositionException {
    if (!equals (p.getDocument ())) throw new InvalidPositionException ("Wrong document");
    int last = p.lineNumber ();
    return last;
  };

  /** Locks the whole document. Useful for multi delete. */
  synchronized void lock () {
    if (locks++ != 0) {
       for (;;) {
        try {
          wait ();
          break;
        } catch (InterruptedException e) {
        };
      };
    };
  };
  
  /** unlocks the document */
  synchronized void unlock () {
    // decrement locks and
    if (--locks != 0) notify ();
    // notify if there is someone waiting
  };

  /** Locks redrawing. Must be followed by the unlock.
  */
  public void redrawLock () {
    Enumeration e = redraw.elements ();
    while (e.hasMoreElements ()) {
      ((RedrawListener)(e.nextElement ())).redrawLock (this);
    };
  };

  /** Unlocks redrawing. Must be called after the lock.
  */
  public void redrawUnlock () {
    Enumeration e = redraw.elements ();
    while (e.hasMoreElements ()) {
      ((RedrawListener)(e.nextElement ())).redrawUnlock (this);
    };
  };

  /** Unlocks redrawing. Must be called after the lock.
  * @param b the block
  */
  public void redrawBlock (Block b) {
    Enumeration e = redraw.elements ();
    while (e.hasMoreElements ()) {
      ((RedrawListener)(e.nextElement ())).redrawBlock (b);
    };
  };

  /** Sends a message to block's document with a request to redraw. If a
  * part of the document is visible, than it should be repainted.
  */
  public void redraw () {
    redrawBlock (this);
  };
  
  /** Sets modified. If the document changed its modification state
  * fires notification to all registered DocumentListeners.
  * @param mod true to indicate modified state
  * @see DocumentListener
  */
  public void setModified (boolean mod) {
    if (modified != mod) {
      modified = mod;
      Enumeration e = doc.elements ();
      while (e.hasMoreElements ()) {
        ((DocumentListener)(e.nextElement ())).documentModified (this, mod);
      };
    };
  };

  /** Returns if the document is modified or not.
  * @return true if modified
  */
  public boolean isModified () {
    return modified;
  };
  
  /** Sends a message to all DocumentListeners on this document.
  * @param s the message
  * @see DocumentListener
  */
  public void message (String s) {
    Enumeration e = doc.elements ();
    while (e.hasMoreElements ()) {
      ((DocumentListener)(e.nextElement ())).documentMessage (this, s);
    };
  };
};



