/**
* Xelfi programming environment
*/

package xelfi.editor;
import java.awt.Event;
/**
* Implements action controler and watches if the related block is not empty.
* If it is the specified text is inserted.
*
* @version 1.0
* @author Jaroslav Tulach
*/

public class ForbidEmptyBlock extends Object implements ActionController {
  /** The block to watch around */
  Block block;

  /** The text to insert */
  String text;
  
  /** Creates new controler. Now the block b becomes guarded and if it ever empties
  * it will be filled with text s.
  *
  * @param b the block to watch over
  * @param s the string to insert if the block is empty
  */
  public ForbidEmptyBlock (Block b, String s) {
    block = b;
    text = s;
    b.getBegin ().getDocument ().addActionController (this);
  };

  /** Creates new controler. If the block becomes empty new line is inserted.
  * @param b the block to watch over
  */
  public ForbidEmptyBlock (Block b) {
    this (b, "\n");
  };

  /** Stops watching over specified block. */
  public void stopWatching () {
    block.getBegin ().getDocument ().removeActionController (this);
  };

  /** This method is called before every insert to any position
  * For multiple inserts (of arrays or Strings) is called only
  * once at begining.
  *
  *
  * @param p the position method insert will insert to
  * @param info optional info sent to insert or null if nothing has been sent
  * @exception NotAllowedException if this insert is not allowed
  */
  public void allowInsert (Position p, Object info) throws NotAllowedException {
  };
  
  /** This method is called before every delete of single character at any position.
  *
  * @param p the position method delete will delete at
  * @param info optional info sent to delete
  * @exception NotAllowedException if this delete is not allowed
  */
  public void allowDelete (Position p, Object info) throws NotAllowedException {
    try {
      Position b = block.getBegin ();

      if (p.equals (b)) {
        Position e = block.getEnd ();
        if (b.next ().equals (e)) {
          // a try to delete the last character in block
          e.insert (text, this);
          // try to insert a specified text, if it falls than
          // do not allow insert (because the NotAllowedException is
          // thrown, otherwise allow delete of the character
          return;
        };
        if (p.isEndOfLine () && e.previousChar ().lessOrEqual (p)) {
          e.insert (text, this);
          // try to insert a specified text, if it falls than
          // do not allow insert (because the NotAllowedException is
          // thrown, otherwise allow delete of the character
          return;
        };
      };
    } catch (EndOfTextException ex) {
      throw new NotAllowedException ();
    };
  };

  /** This method is called before every delete of a block (more characters).
  * @param from the first position to delete
  * @param to the last (excluded) position
  * @param info the info
  * @exception NotAllowedException if this delete is not allowed
  */
  public void allowMultiDelete (Position from, Position to, Object info)
  throws NotAllowedException {
    if (from.lessOrEqual (block.getBegin ()) && to.greaterOrEqual (block.getEnd ())) {
      throw new NotAllowedException ();
    };
  };
  
  /** Notifies that an insert has been done at position p.
  * @param n the begin of inserted text
  * @param p the end of inserted text
  */
  public void notifyInsert (Position n, Position p) {
  };
  
  /** Notifies a delete
  * @param p the position where a delete occures
  * @param s the string with characters that has been deleted
  */
  public void notifyDelete (Position p, String s) {
  };
  
  /** Handle event allows sending some predefined messages to position
  * and by throw the document to ActionController
  * @param p the position a message is sent to
  * @param e the event (the same as in java.awt)
  * @return true to signal successful handling
  * @see java.awt.Event
  */
  public boolean handleEvent (Position p, Event e) {
    return false;
  };
};


