/**
* Xelfi Environment
*/

package xelfi.editor.keys;
import xelfi.editor.*;
import java.util.*;

/** Maps keys to context items.
*
* @author Jaroslav Tulach
* @version 1.0
*/
final class ContextTable extends Object implements ContextItem {
  /** Hashtable that contains
  * ContextItems[KeyCode.STATE_COUNT] indexed by Integers
  */
  Hashtable table = new Hashtable ();

  /** Access right item
  * @param key the key code to access
  * @return the context item for this key or null if no such item
  *   exists
  */
  private ContextItem access (KeyCode key) {
    ContextItem[] ar = (ContextItem[]) table.get (key);
    return ar == null ? null : ar[key.mod];
  }

  /** Executes a key.
  * @param key the key to execute
  * @param info info about the editor
  * @return the context that should be used next time or null if
  *    next time the basic context table should be used
  * @exception NoSuchElementException if no key reaction is found
  * @exception NotAllowedException if action is not allowed
  * @exception EndOfTextException if end of text is reached
  */
  public final ContextTable distributeKey (KeyCode key, TextEditorInfo info)
  throws NoSuchElementException, NotAllowedException, EndOfTextException {
    ContextItem ci = access (key);
    if (ci == null) {
      // no reaction is defined
      throw new NoSuchElementException ();
    }

    return ci.acceptKey (key, info);
  }

  /** Accepts key. For ContextTable this means that it returns itself
  * to be asked for next key.
  * @param key the key
  * @return itself
  */
  public ContextTable acceptKey (KeyCode key, TextEditorInfo info) {
    return this;
  }

  /** Adds new key to the list.
  * @param key the key to add
  * @param v the handler vector that should react to this key action
  */
  public boolean addKey (KeyCode key, KeyActionMacro v) {
    boolean[] sets = key.sets ();
    ContextItem[] current = (ContextItem[]) table.get (key);
    if (current == null) {
      // add new key
      current = new ContextItem[KeyCode.STATE_COUNT];
      table.put (key, current);
      return addNewKey (key, v, sets, current);
    }

    int first = -1;
    for (int i = 0; i < KeyCode.STATE_COUNT; i++) {
      if (sets[i]) {
        // i-position is must be set
        if (first == -1) {
          // this is the first position that must be set
          first = i;
        } else {
          // test if the first position and this points to the same
          // ContextItem
          if (current[first] != current[i]) {
            // they are different
            return false;
          }
          // else go on searching
        }
      }
    }
    if (current[first] != null) {
      // there is already a reaction on this modifiers
      // all reactions are the same
      if (key.next == null) {
        // this cannot be last key, there is sequence of next one(s)
        return false;
      }
      return current[first].addKey (key.next, v);
    } else {
      return addNewKey (key, v, sets, current);
    }
  }

  /** Adds new key.
  * @param key the key
  * @param v the vector
  * @param sets the set that indicates where to set
  * @param current the vector where to set
  */
  private boolean addNewKey (
    KeyCode key, KeyActionMacro v, boolean[] sets, ContextItem[] current
  ) {
    // we must set new ContextItem
    if (key.next == null) {
      ContextItem ci = new ContextAction (v);
      setToTable (current, sets, ci);
      return true;
    } else {
      // there is next sequence of keys, create the ContextTable
      ContextTable ci = new ContextTable ();
      setToTable (current, sets, ci);
      return ci.addKey (key.next, v);
    }
  }

  /** To every position in table, where boolean array is true
  * the value what is assigned.
  * @param table array of ContextItem
  * @param where indicator array
  * @param what what to set
  */
  private void setToTable (ContextItem[] table, boolean[] where, ContextItem what) {
    for (int i = 0; i < table.length; i++) {
      if (where[i]) {
        table[i] = what;
      }
    }
  }

  /** toString.
  * @return string
  */
  public String toString () {
    StringBuffer sb = new StringBuffer ("Table:\n");
    Enumeration en = table.keys ();
    while (en.hasMoreElements ()) {
      KeyCode key = ((KeyCode)en.nextElement ());
      ContextItem[] ci = (ContextItem[]) table.get (key);
      for (int i = 0; i < KeyCode.STATE_COUNT; i++) {
        sb.append ("" + key.key + '[' + i  + "] = ");
        if (ci[i] == null) {
          sb.append ("null\n");
        } else {
          StringTokenizer st = new StringTokenizer (ci[i].toString (), "\n");
          while (st.hasMoreTokens ()) {
            sb.append ("  ");
            sb.append (st.nextToken ());
            sb.append ('\n');
          }
        }
      }
    }
    return sb.toString ();
  }
}


























// Generated from substitution template 'Class' Blocks:Class:12,13,24,13; Package:8,4,25,4; 
