/**
* Xelfi programming environment
*/

package xelfi.templates;
import java.util.NoSuchElementException;
import xelfi.editor.Source;
import java.io.IOException;
/**
* Provides basic functions for manipulation with
* templates.
* Every template has its own instance of TemplateManager.
* But it should be offen that more templates has the same
* kind of template manager (the same class).
*
* This class then behaves differently for each template.
*
* @version $Id$
* @autor Jaroslav Tulach
*/

public interface TemplateManager {
  /** This method is called before a Template informations
  * are saved to string. This method can add any specific
  * information about the template in the string.
  * @param t the template
  * @return the string to add to the specification of template
  */
  public String storeInfo (Template t);

  /** This method does the reverse process as the storeInfo
  * does. It gets the string, reads specific information for
  * this mangager and updates the template t.
  * @param t the template to store information to
  * @param s the string to read parameters from
  * @exception IllegalArgumentException if the parameters in s have wrong format
  */
  public void loadInfo (Template t, String s) throws IllegalArgumentException;

  /** When a new Source should be
  * created the following method
  * is called to insert basic text into the document and/or create
  * guarded blocks.
  * <p>
  * This method should somehow mark the file that it has been
  * created from this template t. So on next open of the source
  * the method identifySource (Source) could identify that
  * it has been created from template t.
  *
  * @param t the template for creating
  * @param s the target source
  * @exception Exception if something went wrong
  */
  public void newSource (Template t, Source s) throws Exception;

  /** This method is called when a new source is loaded and
  * should identify the source's template (the one the source
  * was created from).
  * <p>
  * This method can do any changes to document s (insert text,
  * create guarded blocks) but only if
  * it accepts the source.
  * For infomation how to find out the template if you know
  * only name see Template.find.
  *
  * @param s the source to identify
  * @return the template the source has been created from
  * @exception NoSuchElementException if the template cannot
  *   by recognized by this TemplateManager
  * @exception IOException if the template is recognized but another
  *   error occurs
  *
  * @warning For each TemplateManager registered to the system
  * (it has at least one Template) the method is called only
  * once.
  * @see Template#find
  */
  public Template identifySource (Source s) throws NoSuchElementException, IOException;

  /** When user creates new template from a source this method
  * is called for the template manager that identified the source
  * on its load. This manager should somehow create data structures
  * that are needed for the new template by using the old one.
  * @param old the old template
  * @param s the source
  * @param theNew the new template that should be created
  * @return returns template manager for theNew template (can be this or something else)
  * @exception Exception if the creating is not possible
  */
  public TemplateManager createTemplate (Template old, Source s, Template theNew) throws Exception;
};








