/** Desktop class - setting up and storing the desktop
* @author: Filip Dvorak
* @version: 1.1
*/

package xelfi.top;

import java.io.File;
import java.awt.*;
import java.util.*;

import xelfi.util.*;
import xelfi.design.inspector.XFComponentInspector;
import xelfi.design.components.palette.XFComponentPalette;
import xelfi.top.visual.PaletteWindow;
import xelfi.compiler.Output;
import xelfi.xedit.*;
import xelfi.debugger.*;
import xelfi.browser.view.*;

  /** This class keeps settings of a desktop windows and restores them on startup. */
public class Desktop implements XelfiOptions
{
    // desktop options
  static boolean inspectorShowed = false;    // initialization for the case of no config file
  static Rectangle inspectorPos = null;

  static boolean outputShowed = false;
  static Rectangle outputPos = null;

  static boolean watchesShowed = false;
  static Rectangle watchesPos = null;

  static boolean localsShowed = false;
  static Rectangle localsPos = null;

  static boolean callStackShowed = false;
  static Rectangle callStackPos = null;

  static boolean threadsShowed = false;
  static Rectangle threadsPos = null;

  static boolean breakpointsShowed = false;
  static Rectangle breakpointsPos = null;

  static boolean paletteDocked = true;
  static Rectangle paletteWindow = null;

  static Rectangle mainWinPos = null;  // !! Currently only height and width is used

  static int browserCount = 0;
  static Vector browserId = null;
  static Vector browserShape = null;

    /** Initializes desktop windows(position and visibility)*/
  public static void initDesktop()
  {
    MainMenu menu = TopLevel.getMenu();
    XFComponentPalette pal = (XFComponentPalette)TopLevel.getPalette();
    if(!paletteDocked)
    {
      menu.remove(pal);
      TopLevel.getToolbar().setRows(1);
      menu.show();
      ((CheckboxMenuItem)menu.getMenuItem(MainMenu.miDesignDock)).setState(false);
      new PaletteWindow(menu, pal).show();
    }
    else menu.show();

    if(inspectorShowed) ((XFComponentInspector)TopLevel.getInspector()).show();
    if(outputShowed) TopLevel.getOutput().show();
    if(watchesShowed) TopLevel.getDebugger().getWatches().getView().show();
    if(localsShowed) TopLevel.getDebugger().getLocalsView().show();
    if(callStackShowed) TopLevel.getDebugger().getCallStackView().show();
    if(threadsShowed) TopLevel.getDebugger().getThreadsView().show();
    if(breakpointsShowed) TopLevel.getDebugger().getBreakpoints().getView().show();

    for(int i = 0; i < browserCount; i++)
    {
      Window br = BrowserSettings.CreateBrowserWindow(((Integer)browserId.elementAt(i)).intValue());
      if(br.getPeer() == null) br.addNotify();
      Rectangle r = (Rectangle)browserShape.elementAt(i);
      br.reshape(r.x, r.y, r.width, r.height);
      br.show();
    }
  }

  static void setupWindow(TopWindowBase b)
  {
    Window win = b.getWindow();
    Rectangle menu = TopLevel.getMenu().bounds();
    Dimension scr = Toolkit.getDefaultToolkit().getScreenSize();

    switch(b.id)
    {
      case TopLevel.W_MAINMENU:
        if(win.getPeer() == null) win.addNotify();
        win.move(0, 0);
        win.pack();      // patch - because of proper size in MS windows
        if(mainWinPos != null) win.resize(mainWinPos.width, mainWinPos.height);
        else
          win.resize(Toolkit.getDefaultToolkit().getScreenSize().width, win.preferredSize().height);
        break;
      case TopLevel.W_EDITOR:
        if(win.getPeer() == null) win.addNotify();
        win.move(250, menu.height);
        break;
      case TopLevel.W_INSPECTOR:
        if(win.getPeer() == null) win.addNotify();
        if(inspectorPos != null) win.reshape(inspectorPos.x, inspectorPos.y, inspectorPos.width, inspectorPos.height);
        else win.reshape(0, menu.y + menu.height, 250, 400);
        break;
      case TopLevel.W_OUTPUT:
        if(win.getPeer() == null) win.addNotify();
        if(outputPos != null) win.reshape(outputPos.x, outputPos.y, outputPos.width, outputPos.height);
        else
        {
          int width = scr.width / 2, height = (scr.height - menu.height) / 3;
          win.reshape(scr.width / 2, scr.height - height, width, height);
        }
        break;
      case TopLevel.W_WATCHES:
        if(win.getPeer() == null) win.addNotify();
        if(watchesPos != null) win.reshape(watchesPos.x, watchesPos.y, watchesPos.width, watchesPos.height);
        else
        {
          int width = scr.width / 2, height = (scr.height - menu.height) / 3;
          int dif;
          if((dif = (height + menu.height + 400) - scr.height) > 0) height -= dif;  // 400 = component inspector
          win.reshape(0, scr.height - height, width, height);
        }
        break;
      case TopLevel.W_LOCALS:
        if(win.getPeer() == null) win.addNotify();
        if(localsPos != null) win.reshape(localsPos.x, localsPos.y, localsPos.width, localsPos.height);
        else
        {
          int width = (int)(scr.width * (3f / 12f)), height = (scr.height - menu.height) / 6;
          win.reshape(scr.width - width, menu.height, width, height);
        }
        break;
      case TopLevel.W_CALLSTACK:
        if(win.getPeer() == null) win.addNotify();
        if(callStackPos != null) win.reshape(callStackPos.x, callStackPos.y, callStackPos.width, callStackPos.height);
        else
        {
          int width = (int)(scr.width * (3f / 12f)), height = (scr.height - menu.height) / 6;
          win.reshape(scr.width - width, menu.height + 2 * height, width, height);
        }
        break;
      case TopLevel.W_THREADS:
        if(win.getPeer() == null) win.addNotify();
        if(threadsPos != null) win.reshape(threadsPos.x, threadsPos.y, threadsPos.width, threadsPos.height);
        else
        {
          int width = (int)(scr.width * (3f / 12f)), height = (scr.height - menu.height) / 6;
          win.reshape(scr.width - width, menu.height + 3 * height, width, height);
        }
        break;
      case TopLevel.W_BREAKPOINTS:
        if(win.getPeer() == null) win.addNotify();
        if(breakpointsPos != null) win.reshape(breakpointsPos.x, breakpointsPos.y, breakpointsPos.width, breakpointsPos.height);
        else
        {
          int width = (int)(scr.width * (3f / 12f)), height = (scr.height - menu.height) / 6;
          win.reshape(scr.width - width, menu.height + height, width, height);
        }
        break;
      case TopLevel.W_PALETTEWIN:
        if(win.getPeer() == null) win.addNotify();
        if(paletteWindow != null)
        {
          win.reshape(paletteWindow.x, paletteWindow.y, paletteWindow.width, paletteWindow.height);
          paletteWindow = null;
        }
        break;
      case TopLevel.W_MENUEDITOR:
        if(win.getPeer() == null) win.addNotify();
        win.reshape(0,350,200,400);
      break;
    }
  }

    /** Stores current desktop state into Properties structure that will be later saved into a file.
    @param prop properties to store the desktop */
  public void storeProperties(EnhProperties prop)
  {
    XFComponentInspector insp = (XFComponentInspector)TopLevel.getInspector();
    prop.setBooleanProperty("InspectorVisible", insp.isVisible());
      // are values of position valid ?
    if(inspectorPos != null || !insp.getBase().firstShow)
      if(!insp.getBase().firstShow) prop.setRectProperty("InspectorShape", insp.bounds());
      else prop.setRectProperty("InspectorShape", inspectorPos);

    Output o = TopLevel.getOutput();
    prop.setBooleanProperty("OutputVisible", o.isVisible());
    if(outputPos != null || !o.getBase().firstShow)
      if(!o.getBase().firstShow) prop.setRectProperty("OutputShape", o.bounds());
      else prop.setRectProperty("OutputShape", outputPos);

    WatchesView watch = TopLevel.getDebugger().getWatches().getView();
    prop.setBooleanProperty("WatchesVisible", watch.isVisible());
    if(watchesPos != null || !watch.getBase().firstShow)
      if(!watch.getBase().firstShow) prop.setRectProperty("WatchesShape", watch.bounds());
      else prop.setRectProperty("WatchesShape", watchesPos);

    LocalsView loc = TopLevel.getDebugger().getLocalsView();
    prop.setBooleanProperty("LocalsVisible", loc.isVisible());
    if(localsPos != null || !loc.getBase().firstShow)
      if(!loc.getBase().firstShow) prop.setRectProperty("LocalsShape", loc.bounds());
      else prop.setRectProperty("LocalsShape", localsPos);

    CallStackView stack = TopLevel.getDebugger().getCallStackView();
    prop.setBooleanProperty("CallStackVisible", stack.isVisible());
    if(callStackPos != null || !stack.getBase().firstShow)
      if(!stack.getBase().firstShow) prop.setRectProperty("CallStackShape", stack.bounds());
      else prop.setRectProperty("CallStackShape", callStackPos);

    ThreadsView thr = TopLevel.getDebugger().getThreadsView();
    prop.setBooleanProperty("ThreadsVisible", thr.isVisible());
    if(threadsPos != null || !thr.getBase().firstShow)
      if(!thr.getBase().firstShow) prop.setRectProperty("ThreadsShape", thr.bounds());
      else prop.setRectProperty("ThreadsShape", threadsPos);

    BreakpointsView breaks = TopLevel.getDebugger().getBreakpoints().getView();
    prop.setBooleanProperty("BreakpointsVisible", breaks.isVisible());
    if(breakpointsPos != null || !breaks.getBase().firstShow)
      if(!breaks.getBase().firstShow) prop.setRectProperty("BreakpointsShape", breaks.bounds());
      else prop.setRectProperty("BreakpointsShape", breakpointsPos);

    prop.setRectProperty("MainWindowShape", TopLevel.getMenu().bounds());

    Vector browsers = TopLevel.getBrowsers();
    int size = browsers.size();
    prop.setIntProperty("BrowsersCount", size);
    for(int i = 0; i < size; i++)
    {
     Object b = browsers.elementAt(i);
      prop.setIntProperty("Browser#" + i, ((BrWindow)b).getID());
      prop.setRectProperty("Browser#" + i + "_shape", ((Component)b).bounds());
    }

    Container palwin = ((XFComponentPalette)TopLevel.getPalette()).getParent();

    if(palwin instanceof MainMenu) prop.setBooleanProperty("DockedPalette", true);
    else
    {
      prop.setBooleanProperty("DockedPalette", false);
      prop.setRectProperty("PaletteWindow", palwin.bounds());
    }
  }

    /** Stores desktop state into internal data structures.
    @param prop properties to load the desktop from*/
  public void loadProperties(EnhProperties prop)
  {
      // desktop values
    inspectorShowed = prop.getBooleanProperty("InspectorVisible", false);
    inspectorPos = prop.getRectProperty("InspectorShape", null);

    outputShowed = prop.getBooleanProperty("OutputVisible", false);
    outputPos = prop.getRectProperty("OutputShape", null);

    watchesShowed = prop.getBooleanProperty("WatchesVisible", false);
    watchesPos = prop.getRectProperty("WatchesShape", null);

    localsShowed = prop.getBooleanProperty("LocalsVisible", false);
    localsPos = prop.getRectProperty("LocalsShape", null);

    callStackShowed = prop.getBooleanProperty("CallStackVisible", false);
    callStackPos = prop.getRectProperty("CallStackShape", null);

    threadsShowed = prop.getBooleanProperty("ThreadsVisible", false);
    threadsPos = prop.getRectProperty("ThreadsShape", null);

    breakpointsShowed = prop.getBooleanProperty("BreakpointsVisible", false);
    breakpointsPos = prop.getRectProperty("BreakpointsShape", null);

    paletteDocked = prop.getBooleanProperty("DockedPalette", true);
    paletteWindow = prop.getRectProperty("PaletteWindow", null);

    mainWinPos = prop.getRectProperty("MainWindowShape", mainWinPos);

    if(prop.get("BrowsersCount") != null)
    {
      browserCount = prop.getIntProperty("BrowsersCount", browserCount);
      browserId = new Vector(browserCount);
      browserShape = new Vector(browserCount);
      for(int i = 0; i < browserCount; i++)
      {
        browserId.addElement(new Integer(prop.getIntProperty("Browser#" + i, -1)));
        browserShape.addElement(prop.getRectProperty("Browser#" + i + "_shape", null));
      }
    }
  }

  static void closeAllWindows()
  {
    Enumeration e = TopLevel.allEditors();
    while(e.hasMoreElements())
    {
      JavaEditorFrame ed = (JavaEditorFrame)e.nextElement();
      ed.postEvent(new Event(ed, Event.WINDOW_DESTROY, null));
    }

    e = TopLevel.getBrowsers().elements();
    while(e.hasMoreElements())
    {
      Component c = (Component)e.nextElement();
      c.postEvent(new Event(c, Event.WINDOW_DESTROY, null));
    }

    e = TopLevel.getFormFrames().elements();
    while(e.hasMoreElements())
    {
      Component c = (Component)e.nextElement();
      c.postEvent(new Event(c, Event.WINDOW_DESTROY, null));
    }

    e = TopLevel.getFormDialogs().elements();
    while(e.hasMoreElements())
    {
      Component c = (Component)e.nextElement();
      c.postEvent(new Event(c, Event.WINDOW_DESTROY, null));
    }

    e = TopLevel.getFormApplets().elements();
    while(e.hasMoreElements())
    {
      Component c = (Component)e.nextElement();
      c.postEvent(new Event(c, Event.WINDOW_DESTROY, null));
    }

    TopLevel.getOutput().postEvent(new Event(TopLevel.getOutput(), Event.WINDOW_DESTROY, null));
    ((Component)TopLevel.getMenuEditor()).postEvent(new Event(TopLevel.getMenuEditor(), Event.WINDOW_DESTROY, null));
    ((Component)TopLevel.getInspector()).postEvent(new Event(TopLevel.getInspector(), Event.WINDOW_DESTROY, null));
    TopLevel.getDebugger().getWatches().getView().postEvent(new Event(TopLevel.getDebugger().getWatches().getView(), Event.WINDOW_DESTROY, null));
    TopLevel.getDebugger().getBreakpoints().getView().postEvent(new Event(TopLevel.getDebugger().getBreakpoints().getView(), Event.WINDOW_DESTROY, null));
    TopLevel.getDebugger().getLocalsView().postEvent(new Event(TopLevel.getDebugger().getLocalsView(), Event.WINDOW_DESTROY, null));
    TopLevel.getDebugger().getCallStackView().postEvent(new Event(TopLevel.getDebugger().getCallStackView(), Event.WINDOW_DESTROY, null));
    TopLevel.getDebugger().getThreadsView().postEvent(new Event(TopLevel.getDebugger().getThreadsView(), Event.WINDOW_DESTROY, null));
  }

    /** XelfiOptions implementation.
    @return null as these options are not editable in the options dialog window. */
  public Panel getPanel()
  { return null; }

    /** XelfiOptions implementation.
    @return null as these options are not editable in the options dialog window. */
  public String getTabName()
  { return null; }

    /** XelfiOptions implementation. Empty method. */
  public void onOK(Panel p)
  {}

}



