/**
* Xelfi programming environment
*/

package xelfi.xedit;
import xelfi.editor.*;
import xelfi.templates.*;
import java.util.*;
import java.io.*;
/**
*
* @version 1.0
* @author Jaroslav Tulach
*/

class SubstController extends BlockContainer implements
  ActionController,
  DocumentHandler
{
  /** type of blocks are */
  private static final String PACKAGE = "Package";
  private static final String CLASS = "Class";

  /** hashtable maps blocks to its type */
  Hashtable type = new Hashtable ();

  /** New controler.
  * @param s string with info how to create package and name
  */
  public SubstController (String s, Source so) {
    Document doc = so.getDocument ();
    StringTokenizer st = new StringTokenizer (s, ",;:");
    load (st, so);
    doc.addActionController (this);
  }

  /** This method is called before every insert to any position
  * For multiple inserts (of arrays or Strings) is called only
  * once at begining.
  *
  *
  * @param p the position method insert will insert to
  * @param info optional info sent to insert or null if nothing has been sent
  * @exception NotAllowedException if this insert is not allowed
  */
  public synchronized void allowInsert (Position p, Object info) throws NotAllowedException {
    if (info == this) return;
    try {
      Block b = find (p, Block.INSIDE | Block.END);
      if (!b.isEmpty ()) {
        // insert into block is not allowed
        throw new NotAllowedException ();
      }
    } catch (NoSuchElementException ex) {
    }
  }

  /** This method is called before every delete of single character at any position.
  *
  * @param p the position method delete will delete at
  * @param info optional info sent to delete
  * @exception NotAllowedException if this delete is not allowed
  */
  public synchronized void allowDelete (Position p, Object info) throws NotAllowedException {
    if (info == this) return;
    try {
      Block b = find (p, Block.BEGIN | Block.INSIDE);
      // delete in one of guarded blocks
      // try to delete whole block, if fails then NotAllowedException is thrown
      b.delete ();

      remove (b);
      // in spite of deleting whole block
      // the delete of one character is not allowed
      throw new NotAllowedException (false);
      // forbid the action but do not notify user
    } catch (NoSuchElementException ex) {
    }
  }

  /** This method is called before every delete of a block (more characters).
  * @param from the first position to delete
  * @param to the last (excluded) position
  * @param info the info
  * @exception NotAllowedException if this delete is not allowed
  */
  public synchronized void allowMultiDelete (Position from, Position to, Object info)
  throws NotAllowedException {
    if (info == this) return;
    try {
      find (from, Block.INSIDE);
    } catch (NoSuchElementException ex) {
      try {
        find (to, Block.INSIDE);
      } catch (NoSuchElementException e) {
        // now we know that no block contains from neither to
        // lets allow delete
        return;
      }
    }
    // at least once the block b overlaps with an block in name or pack
    throw new NotAllowedException ();
  }

  /** Generates code the a source.
  * Changes values in all blocks to values that can be
  * obtained from the source. The source must be the
  * same as the one used in function load.
  *
  * @param s the source
  */
  public synchronized void generate (Source s) {
    Enumeration en = elements ();
    String pack = s.getFile ().getPackage ();
    String name = s.getFile ().getName ();
    try {
      for (;;) {
        Block b = (Block) en.nextElement ();
        if (!b.isEmpty ()) {
          String typ = (String) type.get (b);
          try {
            if (typ.equalsIgnoreCase (PACKAGE)) {
              b.set (pack, this);
            } else {
              b.set (name, this);
            }
          } catch (NotAllowedException ex) {
          }
        } else {
          // is empty
          remove (b);
          type.remove (b);
        }
      }
    } catch (NoSuchElementException ex) {
    }
  }

  /** Save BlockContainer.
  * @return the encoding string
  */
  public String save () {
    Enumeration en = elements ();
    StringBuffer sb = new StringBuffer ();
    try {
      Block b = (Block)en.nextElement ();
      for (;;) {
        if (!b.isEmpty ()) {
          String typ = (String) type.get (b);
          sb.append (typ);
          sb.append (':');
          saveBlock (b, sb);
          sb.append (' ');
        }
        b = (Block)en.nextElement ();
      }
    } catch (NoSuchElementException ex) {
    }
    return sb.toString ();
  }

  /** Loads Blockcontainer.
  * @param st the input string tokenizer
  * @param s the source to load from
  */
  public void load (StringTokenizer st, Source s) {
    Document doc = s.getDocument ();
    try {
      for (;;) {
        String typ;
        try {
          typ = st.nextToken ().trim ();
        } catch (NoSuchElementException ex) {
          // this is the last element in the string, so break
          // the for loop
          break;
        }

        Block b = loadBlock (st, doc);
        type.put (b, typ);
        add (b);
      }
    } catch (Exception ex) {
      ex.printStackTrace ();
    }
  }

  /** Adds encoded text for position p
  * @param p the position
  * @param sb the buffer
  */
  static void savePosition (Position p, StringBuffer sb) {
    sb.append (p.getColumn ());
    sb.append (',');
    try {
      sb.append (p.getDocument ().lineToInt (p));
    } catch (InvalidPositionException ex) {
      // should never be thrown
      sb.append ('0');
    }
  }

  /** Loads position from string tokenizer.
  * @param st string tokenizer to read from
  * @param doc to create position in
  * @return new position
  * @exception Exception if an error occures
  */
  static Position loadPosition (StringTokenizer st, Document doc) throws Exception {
    String column = st.nextToken ().trim ();
    String line = st.nextToken ().trim ();
    return doc.goToLine (Integer.valueOf (line).intValue ()).
      column (Integer.valueOf (column).intValue ());
  }

  /** Saves block.
  * @param b the block
  * @param sb string buffer
  */
  static void saveBlock (Block b, StringBuffer sb) {
    savePosition (b.getBegin (), sb);
    sb.append (',');
    savePosition (b.getEnd (), sb);
    sb.append (';');
  }

  /** Loads a block.
  * @param st string tokenizer
  * @param doc the document
  * @return new Block
  * @exception Exception if an error occures
  */
  static Block loadBlock (StringTokenizer st, Document doc) throws Exception {
    Position beg = loadPosition (st, doc);
    Position end = loadPosition (st, doc);
    return new Block (beg, end);
  }

  // methods of DocumentHandler

  /** Calls StandardDocumentHandler.save () to save the source file.
  * @see xelfi.editor.StandardSaver#save
  * @param s the source
  * @param backup is true if the backup files should be created
  * @exception IOException if an error occures and saving is not possible
  */
  public void saveSource (Source s, boolean backup) throws IOException {
    Template t = Template.identifySource (s);

    if (backup) {
      try {
        xelfi.editor.StandardDocumentHandler.backup (s.getSource ());
      } catch (IOException ex) {
      }
    }
    DataOutputStream os = new DataOutputStream (new FileOutputStream (s.getSource ()));
    xelfi.editor.StandardDocumentHandler.save (s, os);
    os.writeBytes (SubstManager.createSignature (t, SubstManager.signatureHeader, SubstManager.signatureEnd));
    os.writeByte (' ');
    os.writeBytes (SubstManager.blocksBegin);
    os.writeBytes (save ());
    os.writeByte ('\n');
    os.close ();
  }
  
  /** Shows the editor. Now calls top level. Bleee....
  * @param s the source
  */
  public void showAll (Source s) {
    xelfi.top.TopLevel.showEditor (s, true);
  }
  
}









